@@ -1,27 +1,25 @@
 from __future__ import print_function, division, absolute_import
 
-import os
-import inspect
 from contextlib import contextmanager
 from collections import namedtuple, defaultdict
-from pprint import pprint
 import sys
 import warnings
 import traceback
-import threading
-from .tracing import trace, event
+from .tracing import event
 
 from numba import (bytecode, interpreter, funcdesc, postproc,
-                   typing, typeinfer, lowering, objmode, utils, config,
-                   errors, types, ir, types, rewrites, transforms)
+                   typing, typeinfer, lowering, pylowering, utils, config,
+                   errors, types, ir, rewrites, transforms)
 from numba.targets import cpu, callconv
 from numba.annotations import type_annotations
-from numba.parfor import ParforPass
+from numba.parfor import PreParforPass, ParforPass, Parfor, ParforDiagnostics
 from numba.inline_closurecall import InlineClosureCallPass
+from numba.errors import CompilerError
+from numba.ir_utils import raise_on_unsupported_feature
+from numba.compiler_lock import global_compiler_lock
 
-
-# Lock for the preventing multiple compiler execution
-lock_compiler = threading.RLock()
+# terminal color markup
+_termcolor = errors.termcolor()
 
 
 class Flags(utils.ConfigOptions):
@@ -44,11 +42,15 @@ class Flags(utils.ConfigOptions):
         'boundcheck': False,
         'forceinline': False,
         'no_cpython_wrapper': False,
-        'auto_parallel': False,
+        # Enable automatic parallel optimization, can be fine-tuned by taking
+        # a dictionary of sub-options instead of a boolean, see parfor.py for
+        # detail.
+        'auto_parallel': cpu.ParallelOptions(False),
         'nrt': False,
         'no_rewrites': False,
         'error_model': 'python',
         'fastmath': False,
+        'noalias': False,
     }
 
 
@@ -69,7 +71,8 @@ CR_FIELDS = ["typing_context",
              "library",
              "call_helper",
              "environment",
-             "has_dynamic_globals"]
+             "has_dynamic_globals",
+             "metadata",]
 
 
 class CompileResult(namedtuple("_CompileResult", CR_FIELDS)):
@@ -108,6 +111,7 @@ class CompileResult(namedtuple("_CompileResult", CR_FIELDS)):
                  typing_error=None,
                  call_helper=None,
                  has_dynamic_globals=False,  # by definition
+                 metadata=None, # Do not store, arbitrary and potentially large!
                  )
         return cr
 
@@ -141,7 +145,8 @@ def compile_result(**kws):
 def compile_isolated(func, args, return_type=None, flags=DEFAULT_FLAGS,
                      locals={}):
     """
-    Compile the function in an isolated environment (typing and target context).
+    Compile the function in an isolated environment (typing and target
+    context).
     Good for testing.
     """
     from .targets.registry import cpu_target
@@ -150,7 +155,7 @@ def compile_isolated(func, args, return_type=None, flags=DEFAULT_FLAGS,
     # Register the contexts in case for nested @jit or @overload calls
     with cpu_target.nested_context(typingctx, targetctx):
         return compile_extra(typingctx, targetctx, func, args, return_type,
-                             flags, locals)
+                            flags, locals)
 
 
 def run_frontend(func):
@@ -219,16 +224,19 @@ class _PipelineManager(object):
         newmsg = "{desc}\n{exc}".format(desc=desc, exc=exc)
 
         # For python2, attach the traceback of the previous exception.
-        if not utils.IS_PY3:
+        if not utils.IS_PY3 and config.FULL_TRACEBACKS:
+            # strip the new message to just print the error string and not
+            # the marked up source etc (this is handled already).
+            stripped = _termcolor.errmsg(newmsg.split('\n')[1])
             fmt = "Caused By:\n{tb}\n{newmsg}"
-            newmsg = fmt.format(tb=traceback.format_exc(), newmsg=newmsg)
+            newmsg = fmt.format(tb=traceback.format_exc(), newmsg=stripped)
 
         exc.args = (newmsg,)
         return exc
 
+    @global_compiler_lock
     def run(self, status):
         assert self._finalized, "PM must be finalized before run()"
-        res = None
         for pipeline_name in self.pipeline_order:
             event(pipeline_name)
             is_final_pipeline = pipeline_name == self.pipeline_order[-1]
@@ -239,7 +247,8 @@ class _PipelineManager(object):
                 except _EarlyPipelineCompletion as e:
                     return e.result
                 except BaseException as e:
-                    msg = "Failed at %s (%s)" % (pipeline_name, stage_name)
+                    msg = "Failed in %s mode pipeline (step: %s)" % \
+                        (pipeline_name, stage_name)
                     patched_exception = self._patch_error(msg, e)
                     # No more fallback pipelines?
                     if is_final_pipeline:
@@ -255,7 +264,7 @@ class _PipelineManager(object):
         raise CompilerError("All pipelines have failed")
 
 
-class Pipeline(object):
+class BasePipeline(object):
     """
     Stores and manages states for the compiler pipeline
     """
@@ -284,6 +293,11 @@ class Pipeline(object):
         self.typemap = None
         self.calltypes = None
         self.type_annotation = None
+        self.metadata = {} # holds arbitrary inter-pipeline stage meta data
+
+        # parfor diagnostics info, add to metadata
+        self.parfor_diagnostics = ParforDiagnostics()
+        self.metadata['parfor_diagnostics'] = self.parfor_diagnostics
 
         self.status = _CompileStatus(
             can_fallback=self.flags.enable_pyobject,
@@ -421,6 +435,25 @@ class Pipeline(object):
                               lifted=tuple(loops), lifted_from=None)
             return cres
 
+    def stage_frontend_withlift(self):
+        """
+        Extract with-contexts
+        """
+        main, withs = transforms.with_lifting(
+            func_ir=self.func_ir,
+            typingctx=self.typingctx,
+            targetctx=self.targetctx,
+            flags=self.flags,
+            locals=self.locals,
+            )
+        if withs:
+            cres = compile_ir(self.typingctx, self.targetctx, main,
+                              self.args, self.return_type,
+                              self.flags, self.locals,
+                              lifted=tuple(withs), lifted_from=None,
+                              pipeline_class=type(self))
+            raise _EarlyPipelineCompletion(cres)
+
     def stage_objectmode_frontend(self):
         """
         Front-end: Analyze bytecode, generate Numba IR, infer types
@@ -444,12 +477,15 @@ class Pipeline(object):
         with self.fallback_context('Function "%s" failed type inference'
                                    % (self.func_id.func_name,)):
             # Type inference
-            self.typemap, self.return_type, self.calltypes = type_inference_stage(
+            typemap, return_type, calltypes = type_inference_stage(
                 self.typingctx,
                 self.func_ir,
                 self.args,
                 self.return_type,
                 self.locals)
+            self.typemap = typemap
+            self.return_type = return_type
+            self.calltypes = calltypes
 
         with self.fallback_context('Function "%s" has invalid return type'
                                    % (self.func_id.func_name,)):
@@ -462,9 +498,10 @@ class Pipeline(object):
         inference.
         """
         assert self.func_ir
-        with self.fallback_context('Internal error in pre-inference rewriting '
-                                   'pass encountered during compilation of '
-                                   'function "%s"' % (self.func_id.func_name,)):
+        msg = ('Internal error in pre-inference rewriting '
+               'pass encountered during compilation of '
+               'function "%s"' % (self.func_id.func_name,))
+        with self.fallback_context(msg):
             rewrites.rewrite_registry.apply('before-inference',
                                             self, self.func_ir)
 
@@ -477,12 +514,29 @@ class Pipeline(object):
         assert self.func_ir
         assert isinstance(getattr(self, 'typemap', None), dict)
         assert isinstance(getattr(self, 'calltypes', None), dict)
-        with self.fallback_context('Internal error in post-inference rewriting '
-                                   'pass encountered during compilation of '
-                                   'function "%s"' % (self.func_id.func_name,)):
+        msg = ('Internal error in post-inference rewriting '
+               'pass encountered during compilation of '
+               'function "%s"' % (self.func_id.func_name,))
+        with self.fallback_context(msg):
             rewrites.rewrite_registry.apply('after-inference',
                                             self, self.func_ir)
 
+    def stage_pre_parfor_pass(self):
+        """
+        Preprocessing for data-parallel computations.
+        """
+        # Ensure we have an IR and type information.
+        assert self.func_ir
+        preparfor_pass = PreParforPass(
+            self.func_ir,
+            self.type_annotation.typemap,
+            self.type_annotation.calltypes, self.typingctx,
+            self.flags.auto_parallel,
+            self.parfor_diagnostics.replaced_fns
+            )
+
+        preparfor_pass.run()
+
     def stage_parfor_pass(self):
         """
         Convert data-parallel computations into Parfor nodes
@@ -490,16 +544,43 @@ class Pipeline(object):
         # Ensure we have an IR and type information.
         assert self.func_ir
         parfor_pass = ParforPass(self.func_ir, self.type_annotation.typemap,
-            self.type_annotation.calltypes, self.return_type)
+            self.type_annotation.calltypes, self.return_type, self.typingctx,
+            self.flags.auto_parallel, self.flags, self.parfor_diagnostics)
         parfor_pass.run()
 
+        if config.WARNINGS:
+            # check the parfor pass worked and warn if it didn't
+            has_parfor = False
+            for blk in self.func_ir.blocks.values():
+                for stmnt in blk.body:
+                    if isinstance(stmnt, Parfor):
+                        has_parfor = True
+                        break
+                else:
+                    continue
+                break
+
+            if not has_parfor:
+                # parfor calls the compiler chain again with a string
+                if not self.func_ir.loc.filename == '<string>':
+                    msg = ("parallel=True was specified but no transformation"
+                           " for parallel execution was possible.")
+                    warnings.warn_explicit(
+                        msg,
+                        errors.NumbaWarning,
+                        self.func_id.filename,
+                        self.func_id.firstlineno
+                        )
+
     def stage_inline_pass(self):
         """
         Inline calls to locally defined closures.
         """
         # Ensure we have an IR and type information.
         assert self.func_ir
-        inline_pass = InlineClosureCallPass(self.func_ir, run_frontend)
+        inline_pass = InlineClosureCallPass(self.func_ir,
+                                            self.flags.auto_parallel,
+                                            self.parfor_diagnostics.replaced_fns)
         inline_pass.run()
         # Remove all Dels, and re-run postproc
         post_proc = postproc.PostProcessor(self.func_ir)
@@ -510,7 +591,6 @@ class Pipeline(object):
             print(("IR DUMP: %s" % name).center(80, "-"))
             self.func_ir.dump()
 
-
     def stage_annotate_type(self):
         """
         Create type annotation after type inference
@@ -533,6 +613,14 @@ class Pipeline(object):
             with open(config.HTML, 'w') as fout:
                 self.type_annotation.html_annotate(fout)
 
+    def stage_dump_diagnostics(self):
+        if self.flags.auto_parallel.enabled:
+            if config.PARALLEL_DIAGNOSTICS:
+                if self.parfor_diagnostics is not None:
+                    self.parfor_diagnostics.dump(config.PARALLEL_DIAGNOSTICS)
+                else:
+                    raise RuntimeError("Diagnostics failed.")
+
     def backend_object_mode(self):
         """
         Object mode compilation
@@ -551,8 +639,9 @@ class Pipeline(object):
 
     def backend_nopython_mode(self):
         """Native mode compilation"""
-        with self.fallback_context("Function %s failed at nopython "
-                                   "mode lowering" % (self.func_id.func_name,)):
+        msg = ("Function %s failed at nopython "
+               "mode lowering" % (self.func_id.func_name,))
+        with self.fallback_context(msg):
             return native_lowering_stage(
                 self.targetctx,
                 self.library,
@@ -560,7 +649,8 @@ class Pipeline(object):
                 self.typemap,
                 self.return_type,
                 self.calltypes,
-                self.flags)
+                self.flags,
+                self.metadata)
 
     def _backend(self, lowerfn, objectmode):
         """
@@ -575,21 +665,23 @@ class Pipeline(object):
 
         lowered = lowerfn()
         signature = typing.signature(self.return_type, *self.args)
-        self.cr = compile_result(typing_context=self.typingctx,
-                                 target_context=self.targetctx,
-                                 entry_point=lowered.cfunc,
-                                 typing_error=self.status.fail_reason,
-                                 type_annotation=self.type_annotation,
-                                 library=self.library,
-                                 call_helper=lowered.call_helper,
-                                 signature=signature,
-                                 objectmode=objectmode,
-                                 interpmode=False,
-                                 lifted=self.lifted,
-                                 fndesc=lowered.fndesc,
-                                 environment=lowered.env,
-                                 has_dynamic_globals=lowered.has_dynamic_globals,
-                                 )
+        self.cr = compile_result(
+            typing_context=self.typingctx,
+            target_context=self.targetctx,
+            entry_point=lowered.cfunc,
+            typing_error=self.status.fail_reason,
+            type_annotation=self.type_annotation,
+            library=self.library,
+            call_helper=lowered.call_helper,
+            signature=signature,
+            objectmode=objectmode,
+            interpmode=False,
+            lifted=self.lifted,
+            fndesc=lowered.fndesc,
+            environment=lowered.env,
+            has_dynamic_globals=lowered.has_dynamic_globals,
+            metadata=self.metadata,
+            )
 
     def stage_objectmode_backend(self):
         """
@@ -601,14 +693,18 @@ class Pipeline(object):
         # Warn if compiled function in object mode and force_pyobject not set
         if not self.flags.force_pyobject:
             if len(self.lifted) > 0:
-                warn_msg = 'Function "%s" was compiled in object mode without forceobj=True, but has lifted loops.' % (self.func_id.func_name,)
+                warn_msg = ('Function "%s" was compiled in object mode without'
+                            ' forceobj=True, but has lifted loops.' %
+                            (self.func_id.func_name,))
             else:
-                warn_msg = 'Function "%s" was compiled in object mode without forceobj=True.' % (self.func_id.func_name,)
+                warn_msg = ('Function "%s" was compiled in object mode without'
+                            ' forceobj=True.' % (self.func_id.func_name,))
             warnings.warn_explicit(warn_msg, errors.NumbaWarning,
                                    self.func_id.filename,
                                    self.func_id.firstlineno)
             if self.flags.release_gil:
-                warn_msg = "Code running in object mode won't allow parallel execution despite nogil=True."
+                warn_msg = ("Code running in object mode won't allow parallel"
+                            " execution despite nogil=True.")
                 warnings.warn_explicit(warn_msg, errors.NumbaWarning,
                                        self.func_id.filename,
                                        self.func_id.firstlineno)
@@ -637,51 +733,113 @@ class Pipeline(object):
                                  lifted=(),
                                  fndesc=None,)
 
+    def stage_ir_legalization(self):
+        raise_on_unsupported_feature(self.func_ir)
+
     def stage_cleanup(self):
         """
         Cleanup intermediate results to release resources.
         """
 
+    def define_pipelines(self, pm):
+        """Child classes override this to customize the pipeline.
+        """
+        raise NotImplementedError()
+
+    def add_preprocessing_stage(self, pm):
+        """Add the preprocessing stage that analyzes the bytecode to prepare
+        the Numba IR.
+        """
+        if self.func_ir is None:
+            pm.add_stage(self.stage_analyze_bytecode, "analyzing bytecode")
+        pm.add_stage(self.stage_process_ir, "processing IR")
+
+    def add_pre_typing_stage(self, pm):
+        """Add any stages that go before type-inference.
+        The current stages contain type-agnostic rewrite passes.
+        """
+        if not self.flags.no_rewrites:
+            if self.status.can_fallback:
+                pm.add_stage(self.stage_preserve_ir,
+                             "preserve IR for fallback")
+            pm.add_stage(self.stage_generic_rewrites, "nopython rewrites")
+        pm.add_stage(self.stage_inline_pass,
+                     "inline calls to locally defined closures")
+
+    def add_typing_stage(self, pm):
+        """Add the type-inference stage necessary for nopython mode.
+        """
+        pm.add_stage(self.stage_nopython_frontend, "nopython frontend")
+        pm.add_stage(self.stage_annotate_type, "annotate type")
+
+    def add_optimization_stage(self, pm):
+        """Add optimization stages.
+        """
+        if self.flags.auto_parallel.enabled:
+            pm.add_stage(self.stage_pre_parfor_pass,
+                         "Preprocessing for parfors")
+        if not self.flags.no_rewrites:
+            pm.add_stage(self.stage_nopython_rewrites, "nopython rewrites")
+        if self.flags.auto_parallel.enabled:
+            pm.add_stage(self.stage_parfor_pass, "convert to parfors")
+
+    def add_lowering_stage(self, pm):
+        """Add the lowering (code-generation) stage for nopython-mode
+        """
+        pm.add_stage(self.stage_nopython_backend, "nopython mode backend")
+
+    def add_cleanup_stage(self, pm):
+        """Add the clean-up stage to remove intermediate results.
+        """
+        pm.add_stage(self.stage_cleanup, "cleanup intermediate results")
+
+    def add_with_handling_stage(self, pm):
+        pm.add_stage(self.stage_frontend_withlift, "Handle with contexts")
+
+    def define_nopython_pipeline(self, pm, name='nopython'):
+        """Add the nopython-mode pipeline to the pipeline manager
+        """
+        pm.create_pipeline(name)
+        self.add_preprocessing_stage(pm)
+        self.add_with_handling_stage(pm)
+        self.add_pre_typing_stage(pm)
+        self.add_typing_stage(pm)
+        self.add_optimization_stage(pm)
+        pm.add_stage(self.stage_ir_legalization,
+                     "ensure IR is legal prior to lowering")
+        self.add_lowering_stage(pm)
+        pm.add_stage(self.stage_dump_diagnostics, "dump diagnostics")
+        self.add_cleanup_stage(pm)
+
+    def define_objectmode_pipeline(self, pm, name='object'):
+        """Add the object-mode pipeline to the pipeline manager
+        """
+        pm.create_pipeline(name)
+        self.add_preprocessing_stage(pm)
+        pm.add_stage(self.stage_objectmode_frontend,
+                     "object mode frontend")
+        pm.add_stage(self.stage_inline_pass,
+                     "inline calls to locally defined closures")
+        pm.add_stage(self.stage_annotate_type, "annotate type")
+        pm.add_stage(self.stage_ir_legalization,
+                     "ensure IR is legal prior to lowering")
+        pm.add_stage(self.stage_objectmode_backend, "object mode backend")
+        self.add_cleanup_stage(pm)
+
+    def define_interpreted_pipeline(self, pm, name="interp"):
+        """Add the interpreted-mode (fallback) pipeline to the pipeline manager
+        """
+        pm.create_pipeline(name)
+        pm.add_stage(self.stage_compile_interp_mode,
+                     "compiling with interpreter mode")
+        self.add_cleanup_stage(pm)
+
     def _compile_core(self):
         """
         Populate and run compiler pipeline
         """
         pm = _PipelineManager()
-
-        if not self.flags.force_pyobject:
-            pm.create_pipeline("nopython")
-            if self.func_ir is None:
-                pm.add_stage(self.stage_analyze_bytecode, "analyzing bytecode")
-            pm.add_stage(self.stage_process_ir, "processing IR")
-            if not self.flags.no_rewrites:
-                if self.status.can_fallback:
-                    pm.add_stage(self.stage_preserve_ir, "preserve IR for fallback")
-                pm.add_stage(self.stage_generic_rewrites, "nopython rewrites")
-            pm.add_stage(self.stage_inline_pass, "inline calls to locally defined closures")
-            pm.add_stage(self.stage_nopython_frontend, "nopython frontend")
-            pm.add_stage(self.stage_annotate_type, "annotate type")
-            if not self.flags.no_rewrites:
-                pm.add_stage(self.stage_nopython_rewrites, "nopython rewrites")
-            if self.flags.auto_parallel:
-                pm.add_stage(self.stage_parfor_pass, "convert to parfors")
-            pm.add_stage(self.stage_nopython_backend, "nopython mode backend")
-            pm.add_stage(self.stage_cleanup, "cleanup intermediate results")
-
-        if self.status.can_fallback or self.flags.force_pyobject:
-            pm.create_pipeline("object")
-            if self.func_ir is None:
-                pm.add_stage(self.stage_analyze_bytecode, "analyzing bytecode")
-            pm.add_stage(self.stage_process_ir, "processing IR")
-            pm.add_stage(self.stage_objectmode_frontend, "object mode frontend")
-            pm.add_stage(self.stage_annotate_type, "annotate type")
-            pm.add_stage(self.stage_objectmode_backend, "object mode backend")
-            pm.add_stage(self.stage_cleanup, "cleanup intermediate results")
-
-        if self.status.can_giveup:
-            pm.create_pipeline("interp")
-            pm.add_stage(self.stage_compile_interp_mode, "compiling with interpreter mode")
-            pm.add_stage(self.stage_cleanup, "cleanup intermediate results")
-
+        self.define_pipelines(pm)
         pm.finalize()
         res = pm.run(self.status)
         if res is not None:
@@ -706,6 +864,18 @@ class Pipeline(object):
         return self._compile_core()
 
 
+class Pipeline(BasePipeline):
+    """The default compiler pipeline
+    """
+    def define_pipelines(self, pm):
+        if not self.flags.force_pyobject:
+            self.define_nopython_pipeline(pm)
+        if self.status.can_fallback or self.flags.force_pyobject:
+            self.define_objectmode_pipeline(pm)
+        if self.status.can_giveup:
+            self.define_interpreted_pipeline(pm)
+
+
 def _make_subtarget(targetctx, flags):
     """
     Make a new target context from the given target context and flags.
@@ -718,7 +888,7 @@ def _make_subtarget(targetctx, flags):
     if flags.nrt:
         subtargetoptions['enable_nrt'] = True
     if flags.auto_parallel:
-        subtargetoptions['auto_parallel'] = True
+        subtargetoptions['auto_parallel'] = flags.auto_parallel
     if flags.fastmath:
         subtargetoptions['enable_fastmath'] = True
     error_model = callconv.create_error_model(flags.error_model, targetctx)
@@ -728,28 +898,45 @@ def _make_subtarget(targetctx, flags):
 
 
 def compile_extra(typingctx, targetctx, func, args, return_type, flags,
-                  locals, library=None):
+                  locals, library=None, pipeline_class=Pipeline):
+    """Compiler entry point
+
+    Parameter
+    ---------
+    typingctx :
+        typing context
+    targetctx :
+        target context
+    func : function
+        the python function to be compiled
+    args : tuple, list
+        argument types
+    return_type :
+        Use ``None`` to indicate void return
+    flags : numba.compiler.Flags
+        compiler flags
+    library : numba.codegen.CodeLibrary
+        Used to store the compiled code.
+        If it is ``None``, a new CodeLibrary is used.
+    pipeline_class : type like numba.compiler.BasePipeline
+        compiler pipeline
     """
-    Args
-    ----
-    - return_type
-        Use ``None`` to indicate
-    """
-    pipeline = Pipeline(typingctx, targetctx, library,
-                        args, return_type, flags, locals)
+    pipeline = pipeline_class(typingctx, targetctx, library,
+                              args, return_type, flags, locals)
     return pipeline.compile_extra(func)
 
 
 def compile_ir(typingctx, targetctx, func_ir, args, return_type, flags,
-               locals, lifted=(), lifted_from=None, library=None):
+               locals, lifted=(), lifted_from=None, library=None,
+               pipeline_class=Pipeline):
     """
     Compile a function with the given IR.
 
     For internal use only.
     """
 
-    pipeline = Pipeline(typingctx, targetctx, library,
-                        args, return_type, flags, locals)
+    pipeline = pipeline_class(typingctx, targetctx, library,
+                              args, return_type, flags, locals)
     return pipeline.compile_ir(func_ir=func_ir, lifted=lifted,
                                lifted_from=lifted_from)
 
@@ -780,7 +967,7 @@ def legalize_return_type(return_type, interp, targetctx):
                     retstmts.append(inst.value.name)
                 elif isinstance(inst, ir.Assign):
                     if (isinstance(inst.value, ir.Expr)
-                        and inst.value.op == 'cast'):
+                            and inst.value.op == 'cast'):
                         caststmts[inst.target.name] = inst.value
                     elif isinstance(inst.value, ir.Arg):
                         argvars.add(inst.target.name)
@@ -790,12 +977,13 @@ def legalize_return_type(return_type, interp, targetctx):
         for var in retstmts:
             cast = caststmts.get(var)
             if cast is None or cast.value.name not in argvars:
-                raise TypeError("Only accept returning of array passed into the "
-                                "function as argument")
+                raise TypeError("Only accept returning of array passed into "
+                                "the function as argument")
 
     elif (isinstance(return_type, types.Function) or
             isinstance(return_type, types.Phantom)):
-        raise TypeError("Can't return function object in nopython mode")
+        msg = "Can't return function object ({}) in nopython mode"
+        raise TypeError(msg.format(return_type))
 
 
 def translate_stage(func_id, bytecode):
@@ -848,20 +1036,22 @@ def type_inference_stage(typingctx, interp, args, return_type, locals={}):
 
 
 def native_lowering_stage(targetctx, library, interp, typemap, restype,
-                          calltypes, flags):
+                          calltypes, flags, metadata):
     # Lowering
     fndesc = funcdesc.PythonFunctionDescriptor.from_specialized_function(
         interp, typemap, restype, calltypes, mangler=targetctx.mangler,
-        inline=flags.forceinline)
+        inline=flags.forceinline, noalias=flags.noalias)
 
-    lower = lowering.Lower(targetctx, library, fndesc, interp)
-    lower.lower()
-    if not flags.no_cpython_wrapper:
-        lower.create_cpython_wrapper(flags.release_gil)
-    env = lower.env
-    call_helper = lower.call_helper
-    has_dynamic_globals = lower.has_dynamic_globals
-    del lower
+    with targetctx.push_code_library(library):
+        lower = lowering.Lower(targetctx, library, fndesc, interp,
+                               metadata=metadata)
+        lower.lower()
+        if not flags.no_cpython_wrapper:
+            lower.create_cpython_wrapper(flags.release_gil)
+        env = lower.env
+        call_helper = lower.call_helper
+        has_dynamic_globals = lower.has_dynamic_globals
+        del lower
 
     if flags.no_compile:
         return _LowerResult(fndesc, call_helper, cfunc=None, env=env,
@@ -877,15 +1067,18 @@ def native_lowering_stage(targetctx, library, interp, typemap, restype,
 
 
 def py_lowering_stage(targetctx, library, interp, flags):
-    fndesc = funcdesc.PythonFunctionDescriptor.from_object_mode_function(interp)
-    lower = objmode.PyLower(targetctx, library, fndesc, interp)
-    lower.lower()
-    if not flags.no_cpython_wrapper:
-        lower.create_cpython_wrapper()
-    env = lower.env
-    call_helper = lower.call_helper
-    has_dynamic_globals = lower.has_dynamic_globals
-    del lower
+    fndesc = funcdesc.PythonFunctionDescriptor.from_object_mode_function(
+        interp
+        )
+    with targetctx.push_code_library(library):
+        lower = pylowering.PyLower(targetctx, library, fndesc, interp)
+        lower.lower()
+        if not flags.no_cpython_wrapper:
+            lower.create_cpython_wrapper()
+        env = lower.env
+        call_helper = lower.call_helper
+        has_dynamic_globals = lower.has_dynamic_globals
+        del lower
 
     if flags.no_compile:
         return _LowerResult(fndesc, call_helper, cfunc=None, env=env,
