@@ -9,7 +9,7 @@ from numba import utils
 
 # List of bytecodes creating a new block in the control flow graph
 # (in addition to explicit jump labels).
-NEW_BLOCKERS = frozenset(['SETUP_LOOP', 'FOR_ITER'])
+NEW_BLOCKERS = frozenset(['SETUP_LOOP', 'FOR_ITER', 'SETUP_WITH'])
 
 
 class CFBlock(object):
@@ -478,6 +478,7 @@ class ControlFlowAnalysis(object):
         self._curblock = None
         self._blockstack = []
         self._loops = []
+        self._withs = []
 
     def iterblocks(self):
         """
@@ -599,6 +600,16 @@ class ControlFlowAnalysis(object):
         self.jump(inst.next)
         self._force_new_block = True
 
+    def op_SETUP_WITH(self, inst):
+        end = inst.get_jump_target()
+        self._blockstack.append(end)
+        self._withs.append((inst.offset, end))
+        # TODO: WithLifting requires the loop entry be its own block.
+        #       Forcing a new block here is the simplest solution for now.
+        #       But, we should consider other less ad-hoc ways.
+        self.jump(inst.next)
+        self._force_new_block = True
+
     def op_POP_BLOCK(self, inst):
         self._blockstack.pop()
 
