@@ -274,8 +274,19 @@ data_layout = {
 default_data_layout = data_layout[tuple.__itemsize__ * 8]
 
 
+try:
+    NVVM_VERSION = NVVM().get_version()
+except:
+    # the CUDA driver may not be present
+    NVVM_VERSION = (0, 0)
+
 # List of supported compute capability in sorted order
-SUPPORTED_CC = (2, 0), (2, 1), (3, 0), (3, 5), (5, 0), (5, 2)
+if NVVM_VERSION < (1, 4):
+    # CUDA 8.0
+    SUPPORTED_CC = (2, 0), (2, 1), (3, 0), (3, 5), (5, 0), (5, 2), (5, 3), (6, 0), (6, 1), (6, 2)
+else:
+    # CUDA 9.0 and later
+    SUPPORTED_CC = (3, 0), (3, 5), (5, 0), (5, 2), (5, 3), (6, 0), (6, 1), (6, 2), (7, 0)
 
 
 def _find_arch(mycc):
@@ -310,14 +321,14 @@ def get_arch_option(major, minor):
 MISSING_LIBDEVICE_MSG = '''
 Please define environment variable NUMBAPRO_LIBDEVICE=/path/to/libdevice
 /path/to/libdevice -- is the path to the directory containing the libdevice.*.bc
-files in the installation of CUDA.  (requires CUDA >=7.5)
+files in the installation of CUDA.  (requires CUDA >=8.0)
 '''
 
 MISSING_LIBDEVICE_FILE_MSG = '''Missing libdevice file for {arch}.
-Please ensure you have package cudatoolkit 7.5.
+Please ensure you have package cudatoolkit 8.0.
 Install package by:
 
-    conda install cudatoolkit=7.5
+    conda install cudatoolkit=8.0
 '''
 
 
@@ -527,6 +538,7 @@ re_getelementptr = re.compile(r"\bgetelementptr\s(?:inbounds )?\(?")
 re_load = re.compile(r"=\s*\bload\s(?:\bvolatile\s)?")
 
 re_call = re.compile(r"(call\s[^@]+\))(\s@)")
+re_range = re.compile(r"\s*!range\s+!\d+")
 
 re_type_tok = re.compile(r"[,{}()[\]]")
 
@@ -534,6 +546,8 @@ re_annotations = re.compile(r"\bnonnull\b")
 
 re_unsupported_keywords = re.compile(r"\b(local_unnamed_addr|writeonly)\b")
 
+re_parenthesized_list = re.compile(r"\((.*)\)")
+
 
 def llvm39_to_34_ir(ir):
     """
@@ -618,6 +632,21 @@ def llvm39_to_34_ir(ir):
             # to "call ty (...)* @foo"
             line = re_call.sub(r"\1*\2", line)
 
+            # no !range metadata on calls
+            line = re_range.sub('', line).rstrip(',')
+
+            if '@llvm.memset' in line:
+                line = re_parenthesized_list.sub(
+                    _replace_llvm_memset_usage,
+                    line,
+                    )
+        if 'declare' in line:
+            if '@llvm.memset' in line:
+                line = re_parenthesized_list.sub(
+                    _replace_llvm_memset_declaration,
+                    line,
+                    )
+
         # Remove unknown annotations
         line = re_annotations.sub('', line)
 
@@ -626,6 +655,29 @@ def llvm39_to_34_ir(ir):
     return '\n'.join(buf)
 
 
+def _replace_llvm_memset_usage(m):
+    """Replace `llvm.memset` usage for llvm7+.
+
+    Used as functor for `re.sub.
+    """
+    params = list(m.group(1).split(','))
+    align = re.search(r'align (\d+)', params[0]).group(1)
+    params.insert(-1, 'i32 {}'.format(align))
+    out = ', '.join(params)
+    return '({})'.format(out)
+
+
+def _replace_llvm_memset_declaration(m):
+    """Replace `llvm.memset` declaration for llvm7+.
+
+    Used as functor for `re.sub.
+    """
+    params = list(m.group(1).split(','))
+    params.insert(-1, 'i32')
+    out = ', '.join(params)
+    return '({})'.format(out)
+
+
 def set_cuda_kernel(lfunc):
     from llvmlite.llvmpy.core import MetaData, MetaDataString, Constant, Type
 
