@@ -1,12 +1,12 @@
 from __future__ import print_function, absolute_import, division
 import numpy as np
 from numba import cuda
-from numba.cuda.testing import unittest
+from numba.cuda.testing import unittest, SerialMixin
 from numba.cuda.testing import skip_on_cudasim
 
 
 @skip_on_cudasim('Simulator does not have definitions attribute')
-class TestCudaAutoJit(unittest.TestCase):
+class TestCudaAutoJit(SerialMixin, unittest.TestCase):
     def test_autojit(self):
         @cuda.autojit
         def what(a, b, c):
