@@ -1,10 +1,10 @@
 from __future__ import print_function, absolute_import
 import numpy as np
 from numba import cuda
-from numba.cuda.testing import unittest
+from numba.cuda.testing import unittest, SerialMixin
 
 
-class TestCudaComplex(unittest.TestCase):
+class TestCudaComplex(SerialMixin, unittest.TestCase):
     def test_cuda_complex_arg(self):
         @cuda.jit('void(complex128[:], complex128)')
         def foo(a, b):
