@@ -4,12 +4,11 @@ import numpy as np
 
 from numba import cuda
 import numba.unittest_support as unittest
-from numba.cuda.testing import skip_on_cudasim
+from numba.cuda.testing import SerialMixin
 
-@skip_on_cudasim('forall API unsupported in the simulator')
-class TestForAll(unittest.TestCase):
+
+class TestForAll(SerialMixin, unittest.TestCase):
     def test_forall_1(self):
-
         @cuda.jit
         def foo(x):
             i = cuda.grid(1)
@@ -19,10 +18,9 @@ class TestForAll(unittest.TestCase):
         arr = np.arange(11)
         orig = arr.copy()
         foo.forall(arr.size)(arr)
-        self.assertTrue(np.all(arr == orig + 1))
+        np.testing.assert_array_almost_equal(arr, orig + 1)
 
     def test_forall_2(self):
-
         @cuda.jit("void(float32, float32[:], float32[:])")
         def bar(a, x, y):
             i = cuda.grid(1)
@@ -34,9 +32,8 @@ class TestForAll(unittest.TestCase):
         oldy = y.copy()
         a = 1.234
         bar.forall(y.size)(a, x, y)
-        self.assertTrue(np.all(y == (a * x + oldy)))
+        np.testing.assert_array_almost_equal(y, a * x + oldy, decimal=3)
 
 
 if __name__ == '__main__':
     unittest.main()
-
