@@ -6,10 +6,10 @@ from __future__ import print_function, absolute_import, division
 
 import numpy as np
 from numba import cuda, float64
-from numba.cuda.testing import unittest
+from numba.cuda.testing import unittest, SerialMixin
 
 
-class TestLang(unittest.TestCase):
+class TestLang(SerialMixin, unittest.TestCase):
     def test_enumerate(self):
         tup = (1., 2.5, 3.)
 
