@@ -10,14 +10,14 @@ from numba.extending import typeof_impl
 from numba.cuda.kernels.transpose import transpose
 from numba.tracing import event
 from numba import SmartArray
-from numba.cuda.testing import skip_on_cudasim
+from numba.cuda.testing import skip_on_cudasim, SerialMixin
 
 @skip_on_cudasim('Simulator does not support Device arrays')
-class TestJIT(unittest.TestCase):
+class TestJIT(SerialMixin, unittest.TestCase):
     """Test handling of numba.SmartArray"""
 
     def test_transpose(self):
-        
+
         # To verify non-redundant data movement run this test with NUMBA_TRACE=1
         a = SmartArray(np.arange(16, dtype=float).reshape(4,4))
         b = SmartArray(where='gpu', shape=(4,4), dtype=float)
