@@ -1,40 +1,69 @@
 from __future__ import print_function, division, absolute_import
 
-from collections import namedtuple
+import weakref
+import time
+from collections import namedtuple, deque
+import operator
 from functools import partial
 
 from llvmlite.llvmpy.core import Constant, Type, Builder
+from llvmlite import ir as llvmir
 
 from . import (_dynfunc, cgutils, config, funcdesc, generators, ir, types,
                typing, utils)
-from .errors import LoweringError, new_error_context
+from .errors import (LoweringError, new_error_context, TypingError,
+                     LiteralTypingError)
 from .targets import removerefctpass
 from .funcdesc import default_mangler
-from . import debuginfo, parfor
+from . import debuginfo
 
 
 class Environment(_dynfunc.Environment):
-    __slots__ = ()
+    """Stores globals and constant pyobjects for runtime.
+
+    It is often needed to convert b/w nopython objects and pyobjects.
+    """
+    __slots__ = ('env_name', '__weakref__')
+    # A weak-value dictionary to store live environment with env_name as the
+    # key.
+    _memo = weakref.WeakValueDictionary()
 
     @classmethod
     def from_fndesc(cls, fndesc):
-        mod = fndesc.lookup_module()
-        return cls(mod.__dict__)
+        try:
+            # Avoid creating new Env
+            return cls._memo[fndesc.env_name]
+        except KeyError:
+            inst = cls(fndesc.lookup_globals())
+            inst.env_name = fndesc.env_name
+            cls._memo[fndesc.env_name] = inst
+            return inst
 
     def __reduce__(self):
-        return _rebuild_env, (self.globals['__name__'], self.consts)
+        return _rebuild_env, (
+            self.globals['__name__'],
+            self.consts,
+            self.env_name
+            )
 
-    def as_pointer(self, targetctx, ptrty=types.pyobject):
-        """
-        Return a constant pointer for the environment object.
-        """
-        ll_addr = targetctx.get_value_type(types.intp)
-        ll_ptr = targetctx.get_value_type(ptrty)
-        envptr = ll_addr(id(self)).inttoptr(ll_ptr)
-        return envptr
+    def __del__(self):
+        if utils is None or utils.IS_PY3:
+            return
+        if _keepalive is None:
+            return
+        _keepalive.append((time.time(), self))
+        if len(_keepalive) > 10:
+            cur = time.time()
+            while _keepalive and cur - _keepalive[0][0] > 1:
+                _keepalive.popleft()
 
 
-def _rebuild_env(modname, consts):
+_keepalive = deque()
+
+
+def _rebuild_env(modname, consts, env_name):
+    if env_name in Environment._memo:
+        return Environment._memo[env_name]
     from . import serialize
     mod = serialize._rebuild_module(modname)
     env = Environment(mod.__dict__)
@@ -49,13 +78,15 @@ class BaseLower(object):
     """
     Lower IR to LLVM
     """
-    def __init__(self, context, library, fndesc, func_ir):
+
+    def __init__(self, context, library, fndesc, func_ir, metadata=None):
         self.library = library
         self.fndesc = fndesc
         self.blocks = utils.SortedMap(utils.iteritems(func_ir.blocks))
         self.func_ir = func_ir
         self.call_conv = context.call_conv
         self.generator_info = func_ir.generator_info
+        self.metadata = metadata
 
         # Initialize LLVM
         self.module = self.library.create_ir_module(self.fndesc.unique_name)
@@ -73,7 +104,8 @@ class BaseLower(object):
         # Specializes the target context as seen inside the Lowerer
         # This adds:
         #  - environment: the python exceution environment
-        self.context = context.subtarget(environment=self.env)
+        self.context = context.subtarget(environment=self.env,
+                                         fndesc=self.fndesc)
 
         # Debuginfo
         dibuildercls = (self.context.DIBuilder
@@ -126,10 +158,21 @@ class BaseLower(object):
         Called before lowering a block.
         """
 
-    def return_exception(self, exc_class, exc_args=None):
-        self.call_conv.return_user_exc(self.builder, exc_class, exc_args)
+    def return_exception(self, exc_class, exc_args=None, loc=None):
+        self.call_conv.return_user_exc(self.builder, exc_class, exc_args,
+                                       loc=loc,
+                                       func_name=self.func_ir.func_id.func_name)
+
+    def emit_environment_object(self):
+        """Emit a pointer to hold the Environment object.
+        """
+        # Define global for the environment and initialize it to NULL
+        envname = self.context.get_env_name(self.fndesc)
+        gvenv = self.context.declare_env_global(self.module, envname)
 
     def lower(self):
+        # Emit the Env into the module
+        self.emit_environment_object()
         if self.generator_info is None:
             self.genlower = None
             self.lower_normal_function(self.fndesc)
@@ -195,7 +238,7 @@ class BaseLower(object):
         self.debug_print("# function begin: {0}".format(
             self.fndesc.unique_name))
         # Lower all blocks
-        for offset, block in self.blocks.items():
+        for offset, block in sorted(self.blocks.items()):
             bb = self.blkmap[offset]
             self.builder.position_at_end(bb)
             self.lower_block(block)
@@ -250,6 +293,9 @@ class BaseLower(object):
         return self.library.has_dynamic_globals
 
 
+# Dictionary mapping instruction class to its lowering function.
+lower_extensions = {}
+
 class Lower(BaseLower):
     GeneratorLower = generators.GeneratorLower
 
@@ -356,10 +402,11 @@ class Lower(BaseLower):
         elif isinstance(inst, ir.StaticRaise):
             self.lower_static_raise(inst)
 
-        elif isinstance(inst, parfor.Parfor):
-            parfor.lower_parfor_parallel(self, inst)
-
         else:
+            for _class, func in lower_extensions.items():
+                if isinstance(inst, _class):
+                    func(self, inst)
+                    return
             raise NotImplementedError(type(inst))
 
     def lower_setitem(self, target_var, index_var, value_var, signature):
@@ -390,9 +437,9 @@ class Lower(BaseLower):
     def lower_static_raise(self, inst):
         if inst.exc_class is None:
             # Reraise
-            self.return_exception(None)
+            self.return_exception(None, loc=self.loc)
         else:
-            self.return_exception(inst.exc_class, inst.exc_args)
+            self.return_exception(inst.exc_class, inst.exc_args, loc=self.loc)
 
     def lower_assign(self, ty, inst):
         value = inst.value
@@ -446,8 +493,16 @@ class Lower(BaseLower):
         # Yield to caller
         val = self.loadvar(inst.value.name)
         typ = self.typeof(inst.value.name)
-        val = self.context.cast(self.builder, val, typ, self.gentype.yield_type)
-        self.call_conv.return_value(self.builder, val)
+
+        # cast the local val to the type yielded
+        yret = self.context.cast(self.builder, val, typ,
+                                    self.gentype.yield_type)
+
+        # get the return repr of yielded value
+        retval = self.context.get_return_value(self.builder, typ, yret)
+
+        # return
+        self.call_conv.return_value(self.builder, retval)
 
         # Resumption point
         y.lower_yield_resume()
@@ -455,6 +510,10 @@ class Lower(BaseLower):
         return self.context.get_constant_generic(self.builder, retty, None)
 
     def lower_binop(self, resty, expr, op):
+        # if op in utils.OPERATORS_TO_BUILTINS:
+        # map operator.the_op => the corresponding types.Function() TODO: is this looks dodgy ...
+        op = self.context.typing_context.resolve_value_type(op)
+
         lhs = expr.lhs
         rhs = expr.rhs
         static_lhs = expr.static_lhs
@@ -477,37 +536,56 @@ class Lower(BaseLower):
         def try_static_impl(tys, args):
             if any(a is ir.UNDEFINED for a in args):
                 return None
-            static_sig = typing.signature(signature.return_type, *tys)
+            try:
+                if isinstance(op, types.Function):
+                    static_sig = op.get_call_type(self.context.typing_context, tys, {})
+                else:
+                    static_sig = typing.signature(signature.return_type, *tys)
+            except TypingError:
+                return None
             try:
                 static_impl = self.context.get_function(op, static_sig)
                 return static_impl(self.builder, args)
             except NotImplementedError:
                 return None
 
-        res = try_static_impl((types.Const(static_lhs), types.Const(static_rhs)),
-                              (static_lhs, static_rhs))
+        res = try_static_impl(
+            (_lit_or_omitted(static_lhs), _lit_or_omitted(static_rhs)),
+            (static_lhs, static_rhs),
+            )
         if res is not None:
             return cast_result(res)
 
-        res = try_static_impl((types.Const(static_lhs), rty),
-                              (static_lhs, rhs))
+        res = try_static_impl(
+            (_lit_or_omitted(static_lhs), rty),
+            (static_lhs, rhs),
+            )
         if res is not None:
             return cast_result(res)
 
-        res = try_static_impl((lty, types.Const(static_rhs)),
-                              (lhs, static_rhs))
+        res = try_static_impl(
+            (lty, _lit_or_omitted(static_rhs)),
+            (lhs, static_rhs),
+            )
         if res is not None:
             return cast_result(res)
 
         # Normal implementation for generic arguments
-        impl = self.context.get_function(op, signature)
+
+        sig = op.get_call_type(self.context.typing_context, signature.args, {})
+        impl = self.context.get_function(op, sig)
         res = impl(self.builder, (lhs, rhs))
         return cast_result(res)
 
     def lower_getitem(self, resty, expr, value, index, signature):
         baseval = self.loadvar(value.name)
         indexval = self.loadvar(index.name)
-        impl = self.context.get_function("getitem", signature)
+        # Get implementation of getitem
+        op = operator.getitem
+        fnop = self.context.typing_context.resolve_value_type(op)
+        fnop.get_call_type(self.context.typing_context, signature.args, {})
+        impl = self.context.get_function(fnop, signature)
+
         argvals = (baseval, indexval)
         argtyps = (self.typeof(value.name),
                    self.typeof(index.name))
@@ -592,7 +670,7 @@ class Lower(BaseLower):
             if i in inst.consts:
                 pyval = inst.consts[i]
                 if isinstance(pyval, str):
-                    pos_tys[i] = types.Const(pyval)
+                    pos_tys[i] = types.literal(pyval)
 
         fixed_sig = typing.signature(sig.return_type, *pos_tys)
         fixed_sig.pysig = sig.pysig
@@ -603,20 +681,83 @@ class Lower(BaseLower):
 
     def lower_call(self, resty, expr):
         signature = self.fndesc.calltypes[expr]
+        self.debug_print("# lower_call: expr = {0}".format(expr))
         if isinstance(signature.return_type, types.Phantom):
             return self.context.get_dummy_value()
 
         if isinstance(expr.func, ir.Intrinsic):
             fnty = expr.func.name
-            argvals = expr.func.args
         else:
             fnty = self.typeof(expr.func.name)
-            argvals = self.fold_call_args(fnty, signature,
-                                          expr.args, expr.vararg, expr.kws)
 
-        if isinstance(fnty, types.ExternalFunction):
+        if isinstance(fnty, types.ObjModeDispatcher):
+            self.init_pyapi()
+            # Acquire the GIL
+            gil_state = self.pyapi.gil_ensure()
+            # Fix types
+            argnames = [a.name for a in expr.args]
+            argtypes = [self.typeof(a) for a in argnames]
+            argvalues = [self.loadvar(a) for a in argnames]
+            for v, ty in zip(argvalues, argtypes):
+                # Because .from_native_value steal the reference
+                self.incref(ty, v)
+
+            argobjs = [self.pyapi.from_native_value(atyp, aval,
+                                                    self.env_manager)
+                       for atyp, aval in zip(argtypes, argvalues)]
+            # Make Call
+            entry_pt = fnty.dispatcher.compile(tuple(argtypes))
+            callee = self.context.add_dynamic_addr(self.builder,
+                                                   id(entry_pt),
+                                                   info="with_objectmode")
+            ret_obj = self.pyapi.call_function_objargs(callee, argobjs)
+            has_exception = cgutils.is_null(self.builder, ret_obj)
+            with self. builder.if_else(has_exception) as (then, orelse):
+                # Handles exception
+                # This branch must exit the function
+                with then:
+                    # Clean arg
+                    for obj in argobjs:
+                        self.pyapi.decref(obj)
+
+                    # Release the GIL
+                    self.pyapi.gil_release(gil_state)
+
+                    # Return and signal exception
+                    self.call_conv.return_exc(self.builder)
+
+                # Handles normal return
+                with orelse:
+                    # Fix output value
+                    native = self.pyapi.to_native_value(
+                        fnty.dispatcher.output_types,
+                        ret_obj,
+                        )
+                    output = native.value
+
+                    # Release objs
+                    self.pyapi.decref(ret_obj)
+                    for obj in argobjs:
+                        self.pyapi.decref(obj)
+
+                    # cleanup output
+                    if callable(native.cleanup):
+                        native.cleanup()
+
+                    # Release the GIL
+                    self.pyapi.gil_release(gil_state)
+
+                    # Error during unboxing
+                    with self.builder.if_then(native.is_error):
+                        self.call_conv.return_exc(self.builder)
+
+                    res = output
+
+        elif isinstance(fnty, types.ExternalFunction):
             # Handle a named external function
             self.debug_print("# external function")
+            argvals = self.fold_call_args(fnty, signature,
+                                          expr.args, expr.vararg, expr.kws)
             fndesc = funcdesc.ExternalFunctionDescriptor(
                 fnty.symbol, fnty.sig.return_type, fnty.sig.args)
             func = self.context.declare_external_function(self.builder.module,
@@ -627,11 +768,15 @@ class Lower(BaseLower):
         elif isinstance(fnty, types.NumbaFunction):
             # Handle a compiled Numba function
             self.debug_print("# calling numba function")
+            argvals = self.fold_call_args(fnty, signature,
+                                          expr.args, expr.vararg, expr.kws)
             res = self.context.call_internal(self.builder, fnty.fndesc,
                                              fnty.sig, argvals)
 
         elif isinstance(fnty, types.ExternalFunctionPointer):
             self.debug_print("# calling external function pointer")
+            argvals = self.fold_call_args(fnty, signature,
+                                          expr.args, expr.vararg, expr.kws)
             # Handle a C function pointer
             pointer = self.loadvar(expr.func.name)
             # If the external function pointer uses libpython
@@ -670,6 +815,8 @@ class Lower(BaseLower):
 
         elif isinstance(fnty, types.RecursiveCall):
             # Recursive call
+            argvals = self.fold_call_args(fnty, signature,
+                                          expr.args, expr.vararg, expr.kws)
             qualprefix = fnty.overloads[signature.args]
             mangler = self.context.mangler or default_mangler
             mangled_name = mangler(qualprefix, signature.args)
@@ -684,6 +831,13 @@ class Lower(BaseLower):
         else:
             # Normal function resolution
             self.debug_print("# calling normal function: {0}".format(fnty))
+            self.debug_print("# signature: {0}".format(signature))
+            if (isinstance(expr.func, ir.Intrinsic) or
+                    isinstance(fnty, types.ObjModeDispatcher)):
+                argvals = expr.func.args
+            else:
+                argvals = self.fold_call_args(fnty, signature,
+                                              expr.args, expr.vararg, expr.kws)
             impl = self.context.get_function(fnty, signature)
             if signature.recvr:
                 # The "self" object is passed as the function object
@@ -692,7 +846,7 @@ class Lower(BaseLower):
                 # Prepend the self reference
                 argvals = [the_self] + list(argvals)
 
-            res = impl(self.builder, argvals)
+            res = impl(self.builder, argvals, self.loc)
 
             libs = getattr(impl, "libs", ())
             for lib in libs:
@@ -710,7 +864,7 @@ class Lower(BaseLower):
                 return self.lower_binop(resty, expr, expr.fn)
             else:
                 # inplace operators on non-mutable types reuse the same
-                # definition as the corresponding copying operators.
+                # definition as the corresponding copying operators.)
                 return self.lower_binop(resty, expr, expr.immutable_fn)
         elif expr.op == 'unary':
             val = self.loadvar(expr.value.name)
@@ -764,7 +918,7 @@ class Lower(BaseLower):
                 ty = ty.type
 
             # If we have a tuple, we needn't do anything
-            # (and we can't iterate over the heterogenous ones).
+            # (and we can't iterate over the heterogeneous ones).
             if isinstance(ty, types.BaseTuple):
                 assert ty == resty
                 self.incref(ty, val)
@@ -787,7 +941,7 @@ class Lower(BaseLower):
                                                     pair, pairty)
                 with cgutils.if_unlikely(self.builder,
                                          self.builder.not_(is_valid)):
-                    self.return_exception(ValueError)
+                    self.return_exception(ValueError, loc=self.loc)
                 item = self.context.pair_first(self.builder,
                                                pair, pairty)
                 tup = self.builder.insert_value(tup, item, i)
@@ -798,7 +952,7 @@ class Lower(BaseLower):
             is_valid = self.context.pair_second(self.builder,
                                                 pair, pairty)
             with cgutils.if_unlikely(self.builder, is_valid):
-                self.return_exception(ValueError)
+                self.return_exception(ValueError, loc=self.loc)
 
             self.decref(ty.iterator_type, iterobj)
             return tup
@@ -828,11 +982,14 @@ class Lower(BaseLower):
 
                 # Cast the attribute type to the expected output type
                 res = self.context.cast(self.builder, res, attrty, resty)
-            return res
+                return res
 
         elif expr.op == "static_getitem":
-            signature = typing.signature(resty, self.typeof(expr.value.name),
-                                         types.Const(expr.index))
+            signature = typing.signature(
+                resty,
+                self.typeof(expr.value.name),
+                _lit_or_omitted(expr.index),
+                )
             try:
                 # Both get_function() and the returned implementation can
                 # raise NotImplementedError if the types aren't supported
@@ -980,3 +1137,13 @@ class Lower(BaseLower):
             return
 
         self.context.nrt.decref(self.builder, typ, val)
+
+
+def _lit_or_omitted(value):
+    """Returns a Literal instance if the type of value is supported;
+    otherwise, return `Omitted(value)`.
+    """
+    try:
+        return types.literal(value)
+    except LiteralTypingError:
+        return types.Omitted(value)
