@@ -6,12 +6,15 @@ import os
 import sys
 
 from llvmlite import ir
+from llvmlite.binding import Linkage
 import llvmlite.llvmpy.core as lc
 
 from numba import cgutils
 from numba.utils import IS_PY3
 from . import llvm_types as lt
 from numba.compiler import compile_extra, Flags
+from numba.compiler_lock import global_compiler_lock
+
 from numba.targets.registry import cpu_target
 from numba.runtime import nrtdynmod
 
@@ -124,12 +127,14 @@ class _ModuleCompiler(object):
         """
         raise NotImplementedError
 
+    @global_compiler_lock
     def _cull_exports(self):
         """Read all the exported functions/modules in the translator
         environment, and join them into a single LLVM module.
         """
         self.exported_function_types = {}
         self.function_environments = {}
+        self.environment_gvs = {}
 
         codegen = self.context.codegen()
         library = codegen.create_library(self.module_name)
@@ -147,10 +152,10 @@ class _ModuleCompiler(object):
 
         for entry in self.export_entries:
             cres = compile_extra(self.typing_context, self.context,
-                                 entry.function,
-                                 entry.signature.args,
-                                 entry.signature.return_type, flags,
-                                 locals={}, library=library)
+                                entry.function,
+                                entry.signature.args,
+                                entry.signature.return_type, flags,
+                                locals={}, library=library)
 
             func_name = cres.fndesc.llvm_func_name
             llvm_func = cres.library.get_function(func_name)
@@ -165,6 +170,7 @@ class _ModuleCompiler(object):
                     cres.fndesc.restype, cres.fndesc.argtypes)
                 self.exported_function_types[entry] = fnty
                 self.function_environments[entry] = cres.environment
+                self.environment_gvs[entry] = cres.fndesc.env_name
             else:
                 llvm_func.name = entry.symbol
                 self.dll_exports.append(entry.symbol)
@@ -178,8 +184,11 @@ class _ModuleCompiler(object):
         library.finalize()
         for fn in library.get_defined_functions():
             if fn.name not in self.dll_exports:
-                fn.visibility = "hidden"
-
+                if fn.linkage in {Linkage.private, Linkage.internal}:
+                    # Private/Internal linkage must have "default" visibility
+                    fn.visibility = "default"
+                else:
+                    fn.visibility = 'hidden'
         return library
 
     def write_llvm_bitcode(self, output, wrap=False, **kws):
@@ -261,17 +270,36 @@ class _ModuleCompiler(object):
                                               env_defs_init)
         return gv.gep([ZERO, ZERO])
 
+    def _emit_envgvs_array(self, llvm_module, builder, pyapi):
+        """
+        Emit an array of Environment pointers that needs to be filled at
+        initialization.
+        """
+        env_setters = []
+        for entry in self.export_entries:
+            envgv_name = self.environment_gvs[entry]
+            gv = self.context.declare_env_global(llvm_module, envgv_name)
+            envgv = gv.bitcast(lt._void_star)
+            env_setters.append(envgv)
+
+        env_setters_init = lc.Constant.array(lt._void_star, env_setters)
+        gv = self.context.insert_unique_const(llvm_module,
+                                              '.module_envgvs',
+                                              env_setters_init)
+        return gv.gep([ZERO, ZERO])
+
     def _emit_module_init_code(self, llvm_module, builder, modobj,
-                               method_array, env_array):
+                               method_array, env_array, envgv_array):
         """
         Emit call to "external" init function, if any.
         """
         if self.external_init_function:
             fnty = ir.FunctionType(lt._int32,
                                    [modobj.type, self.method_def_ptr,
-                                    self.env_def_ptr])
+                                    self.env_def_ptr, envgv_array.type])
             fn = llvm_module.add_function(fnty, self.external_init_function)
-            return builder.call(fn, [modobj, method_array, env_array])
+            return builder.call(fn, [modobj, method_array, env_array,
+                                     envgv_array])
         else:
             return None
 
@@ -330,8 +358,10 @@ class ModuleCompilerPy2(_ModuleCompiler):
                             lc.Constant.int(lt._int32, sys.api_version)))
 
         env_array = self._emit_environment_array(llvm_module, builder, pyapi)
+        envgv_array = self._emit_envgvs_array(llvm_module, builder, pyapi)
+
         self._emit_module_init_code(llvm_module, builder, mod,
-                                    method_array, env_array)
+                                    method_array, env_array, envgv_array)
         # XXX No way to notify failure to caller...
 
         builder.ret_void()
@@ -484,8 +514,9 @@ class ModuleCompilerPy3(_ModuleCompiler):
             builder.ret(NULL.bitcast(mod_init_fn.type.pointee.return_type))
 
         env_array = self._emit_environment_array(llvm_module, builder, pyapi)
+        envgv_array = self._emit_envgvs_array(llvm_module, builder, pyapi)
         ret = self._emit_module_init_code(llvm_module, builder, mod,
-                                          method_array, env_array)
+                                          method_array, env_array, envgv_array)
         if ret is not None:
             with builder.if_then(cgutils.is_not_null(builder, ret)):
                 # Init function errored out
