@@ -5,6 +5,7 @@ import copy
 import os
 import sys
 from itertools import permutations, takewhile
+from contextlib import contextmanager
 
 import numpy as np
 
@@ -13,9 +14,9 @@ import llvmlite.llvmpy.core as lc
 from llvmlite.llvmpy.core import Type, Constant, LLVMException
 import llvmlite.binding as ll
 
-from numba import llvmthreadsafe as llvmts
 from numba import types, utils, cgutils, typing, funcdesc, debuginfo
 from numba import _dynfunc, _helperlib
+from numba.compiler_lock import global_compiler_lock
 from numba.pythonapi import PythonAPI
 from . import arrayobj, builtins, imputils
 from .imputils import (user_function, user_generator,
@@ -23,7 +24,6 @@ from .imputils import (user_function, user_generator,
                        RegistryLoader)
 from numba import datamodel
 
-
 GENERIC_POINTER = Type.pointer(Type.int(8))
 PYOBJECT = GENERIC_POINTER
 void_ptr = GENERIC_POINTER
@@ -220,6 +220,9 @@ class BaseContext(object):
     # python exceution environment
     environment = None
 
+    # the function descriptor
+    fndesc = None
+
     def __init__(self, typing_context):
         _load_global_helpers()
 
@@ -239,6 +242,7 @@ class BaseContext(object):
         self.special_ops = {}
         self.cached_internal_func = {}
         self._pid = None
+        self._codelib_stack = []
 
         self.data_model_manager = datamodel.default_manager
 
@@ -258,7 +262,7 @@ class BaseContext(object):
         # Populate built-in registry
         from . import (arraymath, enumimpl, iterators, linalg, numbers,
                        optional, polynomial, rangeobj, slicing, smartarray,
-                       tupleobj)
+                       tupleobj, gdb_hook)
         try:
             from . import npdatetime
         except NotImplementedError:
@@ -280,6 +284,35 @@ class BaseContext(object):
         """
         return funcdesc.default_mangler(name, types)
 
+    def get_env_name(self, fndesc):
+        """Get the environment name given a FunctionDescriptior.
+
+        Use this instead of the ``fndesc.env_name`` so that the target-context
+        can provide necessary mangling of the symbol to meet ABI requirements.
+        """
+        return fndesc.env_name
+
+    def declare_env_global(self, module, envname):
+        """Declare the Environment pointer as a global of the module.
+
+        The pointer is initialized to NULL.  It must be filled by the runtime
+        with the actual address of the Env before the associated function
+        can be executed.
+
+        Parameters
+        ----------
+        module :
+            The LLVM Module
+        envname : str
+            The name of the global variable.
+        """
+        if envname not in module.globals:
+            gv = llvmir.GlobalVariable(module, cgutils.voidptr_t, name=envname)
+            gv.linkage = 'common'
+            gv.initializer = cgutils.get_null_value(gv.type.pointee)
+
+        return module.globals[envname]
+
     def get_arg_packer(self, fe_args):
         return datamodel.ArgPacker(self.data_model_manager, fe_args)
 
@@ -376,7 +409,7 @@ class BaseContext(object):
     def declare_function(self, module, fndesc):
         fnty = self.call_conv.get_function_type(fndesc.restype, fndesc.argtypes)
         fn = module.get_or_insert_function(fnty, name=fndesc.mangled_name)
-        self.call_conv.decorate_function(fn, fndesc.args, fndesc.argtypes)
+        self.call_conv.decorate_function(fn, fndesc.args, fndesc.argtypes, noalias=fndesc.noalias)
         if fndesc.inline:
             fn.attributes.add('alwaysinline')
         return fn
@@ -399,6 +432,16 @@ class BaseContext(object):
         gv = self.insert_unique_const(mod, name, text)
         return Constant.bitcast(gv, stringtype)
 
+    def insert_const_bytes(self, mod, bytes, name=None):
+        """
+        Insert constant *byte* (a `bytes` object) into module *mod*.
+        """
+        stringtype = GENERIC_POINTER
+        name = ".bytes.%s" % (name or hash(bytes))
+        text = cgutils.make_bytearray(bytes)
+        gv = self.insert_unique_const(mod, name, text)
+        return Constant.bitcast(gv, stringtype)
+
     def insert_unique_const(self, mod, name, val):
         """
         Insert a unique internal constant named *name*, with LLVM value
@@ -456,7 +499,7 @@ class BaseContext(object):
             impl = self._get_constants.find((ty,))
             return impl(self, builder, ty, val)
         except NotImplementedError:
-            raise NotImplementedError("cannot lower constant of type '%s'" % (ty,))
+            raise NotImplementedError("Cannot lower constant of type '%s'" % (ty,))
 
     def get_constant(self, ty, val):
         """
@@ -479,6 +522,7 @@ class BaseContext(object):
         Return the implementation of function *fn* for signature *sig*.
         The return value is a callable with the signature (builder, args).
         """
+        assert sig is not None
         sig = sig.as_function()
         if isinstance(fn, (types.Function, types.BoundFunction,
                            types.Dispatcher)):
@@ -621,14 +665,14 @@ class BaseContext(object):
 
     def pair_first(self, builder, val, ty):
         """
-        Extract the first element of a heterogenous pair.
+        Extract the first element of a heterogeneous pair.
         """
         pair = self.make_helper(builder, ty, val)
         return pair.first
 
     def pair_second(self, builder, val, ty):
         """
-        Extract the second element of a heterogenous pair.
+        Extract the second element of a heterogeneous pair.
         """
         pair = self.make_helper(builder, ty, val)
         return pair.second
@@ -745,7 +789,8 @@ class BaseContext(object):
     def get_dummy_type(self):
         return GENERIC_POINTER
 
-    def compile_subroutine_no_cache(self, builder, impl, sig, locals={}, flags=None):
+    def _compile_subroutine_no_cache(self, builder, impl, sig, locals={},
+                                     flags=None):
         """
         Invoke the compiler to compile a function to be used inside a
         nopython function, but without generating code to call that
@@ -756,39 +801,53 @@ class BaseContext(object):
         # Compile
         from numba import compiler
 
-        codegen = self.codegen()
-        library = codegen.create_library(impl.__name__)
-        if flags is None:
-            flags = compiler.Flags()
-        flags.set('no_compile')
-        flags.set('no_cpython_wrapper')
-        cres = compiler.compile_internal(self.typing_context, self,
-                                         library,
-                                         impl, sig.args,
-                                         sig.return_type, flags,
-                                         locals=locals)
+        with global_compiler_lock:
+            codegen = self.codegen()
+            library = codegen.create_library(impl.__name__)
+            if flags is None:
+                flags = compiler.Flags()
+            flags.set('no_compile')
+            flags.set('no_cpython_wrapper')
+            cres = compiler.compile_internal(self.typing_context, self,
+                                            library,
+                                            impl, sig.args,
+                                            sig.return_type, flags,
+                                            locals=locals)
 
-        # Allow inlining the function inside callers.
-        codegen.add_linking_library(cres.library)
-        return cres
+            # Allow inlining the function inside callers.
+            codegen.add_linking_library(cres.library)
+            return cres
 
-    def compile_subroutine(self, builder, impl, sig, locals={}):
+    def compile_subroutine(self, builder, impl, sig, locals={}, flags=None,
+                           caching=True):
         """
         Compile the function *impl* for the given *sig* (in nopython mode).
         Return a placeholder object that's callable from another Numba
         function.
+
+        If *caching* evaluates True, the function keeps the compiled function
+        for reuse in *.cached_internal_func*.
         """
         cache_key = (impl.__code__, sig, type(self.error_model))
-        if impl.__closure__:
-            # XXX This obviously won't work if a cell's value is
-            # unhashable.
-            cache_key += tuple(c.cell_contents for c in impl.__closure__)
-        ty = self.cached_internal_func.get(cache_key)
-        if ty is None:
-            cres = self.compile_subroutine_no_cache(builder, impl, sig,
-                                                    locals=locals)
+        if not caching:
+            cached = None
+        else:
+            if impl.__closure__:
+                # XXX This obviously won't work if a cell's value is
+                # unhashable.
+                cache_key += tuple(c.cell_contents for c in impl.__closure__)
+            cached = self.cached_internal_func.get(cache_key)
+        if cached is None:
+            cres = self._compile_subroutine_no_cache(builder, impl, sig,
+                                                     locals=locals,
+                                                     flags=flags)
+            lib = cres.library
             ty = types.NumbaFunction(cres.fndesc, sig)
-            self.cached_internal_func[cache_key] = ty
+            self.cached_internal_func[cache_key] = ty, lib
+
+        ty, lib = self.cached_internal_func[cache_key]
+        # Allow inlining the function inside callers.
+        self.active_code_library.add_linking_library(lib)
         return ty
 
     def compile_internal(self, builder, impl, sig, args, locals={}):
@@ -812,6 +871,8 @@ class BaseContext(object):
 
         with cgutils.if_unlikely(builder, status.is_error):
             self.call_conv.return_status_propagate(builder, status)
+
+        res = imputils.fix_returning_optional(self, builder, sig, status, res)
         return res
 
     def call_unresolved(self, builder, name, sig, args):
@@ -854,6 +915,8 @@ class BaseContext(object):
                                                    sig.args, args)
         with cgutils.if_unlikely(builder, status.is_error):
             self.call_conv.return_status_propagate(builder, status)
+
+        res = imputils.fix_returning_optional(self, builder, sig, status, res)
         return res
 
     def get_executable(self, func, fndesc):
@@ -998,7 +1061,6 @@ class BaseContext(object):
         gv.initializer = addr
         return builder.load(gv)
 
-    @llvmts.lock_llvm
     def get_abi_sizeof(self, ty):
         """
         Get the ABI size of LLVM type *ty*.
@@ -1029,6 +1091,21 @@ class BaseContext(object):
         """
         return lc.Module(name)
 
+    @property
+    def active_code_library(self):
+        """Get the active code library
+        """
+        return self._codelib_stack[-1]
+
+    @contextmanager
+    def push_code_library(self, lib):
+        """Push the active code library for the context
+        """
+        self._codelib_stack.append(lib)
+        try:
+            yield
+        finally:
+            self._codelib_stack.pop()
 
 
 class _wrap_impl(object):
@@ -1039,15 +1116,48 @@ class _wrap_impl(object):
     """
 
     def __init__(self, imp, context, sig):
-        self._imp = imp
+        self._imp = _wrap_missing_loc(imp)
         self._context = context
         self._sig = sig
 
-    def __call__(self, builder, args):
-        return self._imp(self._context, builder, self._sig, args)
+    def __call__(self, builder, args, loc=None):
+        return self._imp(self._context, builder, self._sig, args, loc=loc)
 
     def __getattr__(self, item):
         return getattr(self._imp, item)
 
     def __repr__(self):
         return "<wrapped %s>" % self._imp
+
+
+def _has_loc(fn):
+    """Does function *fn* take ``loc`` argument?
+    """
+    sig = utils.pysignature(fn)
+    return 'loc' in sig.parameters
+
+
+def _wrap_missing_loc(fn):
+    """Wrap function for missing ``loc`` keyword argument.
+    Otherwise, return the original *fn*.
+    """
+    if not _has_loc(fn):
+        def wrapper(*args, **kwargs):
+            kwargs.pop('loc')     # drop unused loc
+            return fn(*args, **kwargs)
+
+        # Copy the following attributes from the wrapped.
+        # Following similar implementation as functools.wraps but
+        # ignore attributes if not available (i.e fix py2.7)
+        attrs = '__name__', 'libs'
+        for attr in attrs:
+            try:
+                val = getattr(fn, attr)
+            except AttributeError:
+                pass
+            else:
+                setattr(wrapper, attr, val)
+
+        return wrapper
+    else:
+        return fn
