@@ -36,7 +36,13 @@ nrt_flags = Flags()
 nrt_flags.set("nrt")
 
 
-tag = testing.make_tag_decorator(['important'])
+tag = testing.make_tag_decorator(['important', 'long_running'])
+
+_windows_py27 = (sys.platform.startswith('win32') and
+                 sys.version_info[:2] == (2, 7))
+_32bit = sys.maxsize <= 2 ** 32
+_reason = 'parfors not supported'
+skip_parfors_unsupported = unittest.skipIf(_32bit or _windows_py27, _reason)
 
 
 class CompilationCache(object):
@@ -407,6 +413,13 @@ class TestCase(unittest.TestCase):
         self.assertPreciseEqual(got, expected)
         return got, expected
 
+
+class SerialMixin(object):
+    """Mixin to mark test for serial execution.
+    """
+    _numba_parallel_test_ = False
+
+
 # Various helpers
 
 @contextlib.contextmanager
@@ -424,6 +437,29 @@ def override_config(name, value):
         setattr(config, name, old_value)
 
 
+@contextlib.contextmanager
+def override_env_config(name, value):
+    """
+    Return a context manager that temporarily sets an Numba config environment
+    *name* to *value*.
+    """
+    old = os.environ.get(name)
+    os.environ[name] = value
+    config.reload_config()
+
+    try:
+        yield
+    finally:
+        if old is None:
+            # If it wasn't set originally, delete the environ var
+            del os.environ[name]
+        else:
+            # Otherwise, restore to the old value
+            os.environ[name] = old
+        # Always reload config
+        config.reload_config()
+
+
 def compile_function(name, code, globs):
     """
     Given a *code* string, compile it with globals *globs* and return
@@ -560,6 +596,13 @@ def captured_stderr():
     return captured_output("stderr")
 
 
+@contextlib.contextmanager
+def capture_cache_log():
+    with captured_stdout() as out:
+        with override_config('DEBUG_CACHE', True):
+            yield out
+
+
 class MemoryLeak(object):
 
     __enable_leak_check = True
