@@ -8,7 +8,7 @@ import numpy as np
 
 import numba.unittest_support as unittest
 from numba.compiler import compile_isolated, Flags
-from numba import jit, typeof, errors, types, utils
+from numba import jit, typeof, errors, types, utils, config
 from .support import TestCase, tag
 
 
@@ -74,6 +74,12 @@ def enumerate_start_usecase():
         result += i * j
     return result
 
+def enumerate_invalid_start_usecase():
+    result = 0
+    for i, j in enumerate((1., 2.5, 3.), 3.14159):
+        result += i * j
+    return result
+
 def filter_usecase(x, filter_func):
     return filter(filter_func, x)
 
@@ -415,6 +421,25 @@ class TestBuiltins(TestCase):
     def test_enumerate_start_npm(self):
         self.test_enumerate_start(flags=no_pyobj_flags)
 
+    def test_enumerate_start_invalid_start_type(self):
+        pyfunc = enumerate_invalid_start_usecase
+        cr = compile_isolated(pyfunc, (), flags=enable_pyobj_flags)
+        with self.assertRaises(TypeError) as raises:
+            cr.entry_point()
+        if config.PYVERSION == (2, 7):
+            thing = 'index'
+        else:
+            thing = 'integer'
+        msg = "'float' object cannot be interpreted as an %s" % thing
+        self.assertIn(msg, str(raises.exception))
+
+    def test_enumerate_start_invalid_start_type_npm(self):
+        pyfunc = enumerate_invalid_start_usecase
+        with self.assertRaises(errors.TypingError) as raises:
+            cr = compile_isolated(pyfunc, (), flags=no_pyobj_flags)
+        msg = "Only integers supported as start value in enumerate"
+        self.assertIn(msg, str(raises.exception))
+
     def test_filter(self, flags=enable_pyobj_flags):
         pyfunc = filter_usecase
         cr = compile_isolated(pyfunc, (types.Dummy('list'),
@@ -695,7 +720,7 @@ class TestBuiltins(TestCase):
         cfunc(1, [1])
 
     def test_max_1_invalid_types(self):
-        # Heterogenous ordering is valid in Python 2
+        # Heterogeneous ordering is valid in Python 2
         if utils.IS_PY3:
             with self.assertRaises(TypeError):
                 self.check_min_max_invalid_types(max_usecase1)
@@ -707,7 +732,7 @@ class TestBuiltins(TestCase):
             self.check_min_max_invalid_types(max_usecase1, flags=no_pyobj_flags)
 
     def test_min_1_invalid_types(self):
-        # Heterogenous ordering is valid in Python 2
+        # Heterogeneous ordering is valid in Python 2
         if utils.IS_PY3:
             with self.assertRaises(TypeError):
                 self.check_min_max_invalid_types(min_usecase1)
