@@ -1,14 +1,19 @@
 from __future__ import print_function, division, absolute_import
 
 import math
+import operator
 import sys
 import pickle
 import multiprocessing
+import ctypes
+from distutils.version import LooseVersion
+import re
 
 import numpy as np
 
 from numba import unittest_support as unittest
-from numba import jit, types, errors, typing
+from numba import jit, types, errors, typing, compiler
+from numba.targets.registry import cpu_target
 from numba.compiler import compile_isolated
 from .support import (TestCase, captured_stdout, tag, temp_directory,
                       override_config)
@@ -22,13 +27,23 @@ from numba.extending import (typeof_impl, type_callable,
                              make_attribute_wrapper,
                              intrinsic, _Intrinsic,
                              register_jitable,
+                             get_cython_function_address
                              )
 from numba.typing.templates import (
-    ConcreteTemplate, signature, infer)
+    ConcreteTemplate, signature, infer, infer_global, AbstractTemplate)
 
 # Pandas-like API implementation
 from .pdlike_usecase import Index, Series
 
+try:
+    import scipy
+    if LooseVersion(scipy.__version__) < '0.19':
+        sc = None
+    else:
+        import scipy.special.cython_special as sc
+except ImportError:
+    sc = None
+
 
 # -----------------------------------------------------------------------
 # Define a custom type and an implicit cast on it
@@ -37,7 +52,6 @@ class MyDummy(object):
     pass
 
 class MyDummyType(types.Opaque):
-
     def can_convert_to(self, context, toty):
         if isinstance(toty, types.Number):
             from numba.typeconv import Conversion
@@ -67,6 +81,33 @@ def unbox_index(typ, obj, c):
     return NativeValue(c.context.get_dummy_value())
 
 
+# -----------------------------------------------------------------------
+# Define a second custom type but w/o implicit cast to Number
+
+class MyDummy2(object):
+    pass
+
+class MyDummyType2(types.Opaque):
+    pass
+
+mydummy_type_2 = MyDummyType2('mydummy_2')
+mydummy_2 = MyDummy2()
+
+@typeof_impl.register(MyDummy2)
+def typeof_mydummy(val, c):
+    return mydummy_type_2
+
+
+def get_dummy_2():
+    return mydummy_2
+
+register_model(MyDummyType2)(models.OpaqueModel)
+
+@unbox(MyDummyType2)
+def unbox_index(typ, obj, c):
+    return NativeValue(c.context.get_dummy_value())
+
+
 # -----------------------------------------------------------------------
 # Define a function's typing and implementation using the classical
 # two-step API
@@ -209,6 +250,54 @@ def overload_len_dummy(arg):
         return len_impl
 
 
+@overload(operator.add)
+def overload_add_dummy(arg1, arg2):
+    if isinstance(arg1, (MyDummyType, MyDummyType2)) and isinstance(arg2, (MyDummyType, MyDummyType2)):
+        def dummy_add_impl(arg1, arg2):
+            return 42
+
+        return dummy_add_impl
+
+
+@overload(operator.getitem)
+def overload_dummy_getitem(obj, idx):
+    if isinstance(obj, MyDummyType) and isinstance(idx, types.Integer):
+        def dummy_getitem_impl(obj, idx):
+            return idx + 123
+        return dummy_getitem_impl
+
+
+def call_add_operator(arg1, arg2):
+    return operator.add(arg1, arg2)
+
+
+def call_add_binop(arg1, arg2):
+    return arg1 + arg2
+
+
+@overload(operator.iadd)
+def overload_iadd_dummy(arg1, arg2):
+    if isinstance(arg1, (MyDummyType, MyDummyType2)) and isinstance(arg2, (MyDummyType, MyDummyType2)):
+        def dummy_iadd_impl(arg1, arg2):
+            return 42
+
+        return dummy_iadd_impl
+
+
+def call_iadd_operator(arg1, arg2):
+    return operator.add(arg1, arg2)
+
+
+def call_iadd_binop(arg1, arg2):
+    arg1 += arg2
+
+    return arg1
+
+
+def call_getitem(obj, idx):
+    return obj[idx]
+
+
 @overload_method(MyDummyType, 'length')
 def overload_method_length(arg):
     def imp(arg):
@@ -276,6 +365,21 @@ def non_boxable_bad_usecase():
     return return_non_boxable()
 
 
+def mk_func_input(f):
+    pass
+
+
+@infer_global(mk_func_input)
+class MkFuncTyping(AbstractTemplate):
+    def generic(self, args, kws):
+        assert isinstance(args[0], types.MakeFunctionLiteral)
+        return signature(types.none, *args)
+
+
+def mk_func_test_impl():
+    mk_func_input(lambda a: a)
+
+
 class TestLowLevelExtending(TestCase):
     """
     Test the low-level two-tier extension API.
@@ -306,6 +410,17 @@ class TestLowLevelExtending(TestCase):
         cr = compile_isolated(pyfunc, (), types.float64)
         self.assertPreciseEqual(cr.entry_point(), 42.0)
 
+    def test_mk_func_literal(self):
+        """make sure make_function is passed to typer class as a literal
+        """
+        test_ir = compiler.run_frontend(mk_func_test_impl)
+        typingctx = cpu_target.typing_context
+        typingctx.refresh()
+        typemap, _, _ = compiler.type_inference_stage(
+            typingctx, test_ir, (), None)
+        self.assertTrue(any(isinstance(a, types.MakeFunctionLiteral)
+                            for a in typemap.values()))
+
 
 class TestPandasLike(TestCase):
     """
@@ -453,6 +568,63 @@ class TestHighLevelExtending(TestCase):
             cfunc(MyDummy())
             self.assertEqual(sys.stdout.getvalue(), "hello!\n")
 
+    def test_add_operator(self):
+        """
+        Test re-implementing operator.add() for a custom type with @overload.
+        """
+        pyfunc = call_add_operator
+        cfunc = jit(nopython=True)(pyfunc)
+
+        self.assertPreciseEqual(cfunc(1, 2), 3)
+        self.assertPreciseEqual(cfunc(MyDummy2(), MyDummy2()), 42)
+
+        # this will call add(Number, Number) as MyDummy implicitly casts to Number
+        self.assertPreciseEqual(cfunc(MyDummy(), MyDummy()), 84)
+
+    def test_add_binop(self):
+        """
+        Test re-implementing '+' for a custom type via @overload(operator.add).
+        """
+        pyfunc = call_add_binop
+        cfunc = jit(nopython=True)(pyfunc)
+
+        self.assertPreciseEqual(cfunc(1, 2), 3)
+        self.assertPreciseEqual(cfunc(MyDummy2(), MyDummy2()), 42)
+
+        # this will call add(Number, Number) as MyDummy implicitly casts to Number
+        self.assertPreciseEqual(cfunc(MyDummy(), MyDummy()), 84)
+
+    def test_iadd_operator(self):
+        """
+        Test re-implementing operator.add() for a custom type with @overload.
+        """
+        pyfunc = call_iadd_operator
+        cfunc = jit(nopython=True)(pyfunc)
+
+        self.assertPreciseEqual(cfunc(1, 2), 3)
+        self.assertPreciseEqual(cfunc(MyDummy2(), MyDummy2()), 42)
+
+        # this will call add(Number, Number) as MyDummy implicitly casts to Number
+        self.assertPreciseEqual(cfunc(MyDummy(), MyDummy()), 84)
+
+    def test_iadd_binop(self):
+        """
+        Test re-implementing '+' for a custom type via @overload(operator.add).
+        """
+        pyfunc = call_iadd_binop
+        cfunc = jit(nopython=True)(pyfunc)
+
+        self.assertPreciseEqual(cfunc(1, 2), 3)
+        self.assertPreciseEqual(cfunc(MyDummy2(), MyDummy2()), 42)
+
+        # this will call add(Number, Number) as MyDummy implicitly casts to Number
+        self.assertPreciseEqual(cfunc(MyDummy(), MyDummy()), 84)
+
+    def test_getitem(self):
+        pyfunc = call_getitem
+        cfunc = jit(nopython=True)(pyfunc)
+        self.assertPreciseEqual(cfunc(MyDummy(), 321), 321 + 123)
+
     def test_no_cpython_wrapper(self):
         """
         Test overloading whose return value cannot be represented in CPython.
@@ -466,7 +638,7 @@ class TestHighLevelExtending(TestCase):
         np.testing.assert_equal(expect, got)
         # Verify that the Module type cannot be returned to CPython
         bad_cfunc = jit(nopython=True)(non_boxable_bad_usecase)
-        with self.assertRaises(NotImplementedError) as raises:
+        with self.assertRaises(TypeError) as raises:
             bad_cfunc()
         errmsg = str(raises.exception)
         expectmsg = "cannot convert native Module"
@@ -589,7 +761,6 @@ class TestIntrinsic(TestCase):
 
         # Test
         arr = np.arange(10, dtype=np.float32)
-        foo(arr)
         with captured_stdout() as buf:
             foo(arr)
             got = buf.getvalue().splitlines()
@@ -664,7 +835,13 @@ class TestIntrinsic(TestCase):
         memo_size += 1
         self.assertEqual(memo_size, len(memo))
         del original  # remove original before unpickling
-        # by deleting, the memo entry is removed
+
+        # by deleting, the memo entry is NOT removed due to recent
+        # function queue
+        self.assertEqual(memo_size, len(memo))
+
+        # Manually force clear of _recent queue
+        _Intrinsic._recent.clear()
         memo_size -= 1
         self.assertEqual(memo_size, len(memo))
 
@@ -709,5 +886,30 @@ class TestRegisterJitable(unittest.TestCase):
         self.assertIn(msg, str(raises.exception))
 
 
+class TestImportCythonFunction(unittest.TestCase):
+    @unittest.skipIf(sc is None, "Only run if SciPy >= 0.19 is installed")
+    def test_getting_function(self):
+        addr = get_cython_function_address("scipy.special.cython_special", "j0")
+        functype = ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double)
+        _j0 = functype(addr)
+        j0 = jit(nopython=True)(lambda x: _j0(x))
+        self.assertEqual(j0(0), 1)
+
+    def test_missing_module(self):
+        with self.assertRaises(ImportError) as raises:
+            addr = get_cython_function_address("fakemodule", "fakefunction")
+        # The quotes are not there in Python 2
+        msg = "No module named '?fakemodule'?"
+        match = re.match(msg, str(raises.exception))
+        self.assertIsNotNone(match)
+
+    @unittest.skipIf(sc is None, "Only run if SciPy >= 0.19 is installed")
+    def test_missing_function(self):
+        with self.assertRaises(ValueError) as raises:
+            addr = get_cython_function_address("scipy.special.cython_special", "foo")
+        msg = "No function 'foo' found in __pyx_capi__ of 'scipy.special.cython_special'"
+        self.assertEqual(msg, str(raises.exception))
+
+
 if __name__ == '__main__':
     unittest.main()
