@@ -220,6 +220,13 @@ class TestJitClass(TestCase, MemoryLeakMixin):
                 # use deferred type as argument
                 return get_data(self.next)
 
+            def append_to_tail(self, other):
+                cur = self
+                while cur.next is not None:
+                    cur = cur.next
+                cur.next = other
+
+
         node_type.define(LinkedNode.class_type.instance_type)
 
         first = LinkedNode(123, None)
@@ -239,6 +246,16 @@ class TestJitClass(TestCase, MemoryLeakMixin):
         first_val = second.get_next_data()
         self.assertEqual(first_val, first.data)
 
+        # Check setattr (issue #2606)
+        self.assertIsNone(first.next)
+        second.append_to_tail(LinkedNode(567, None))
+        self.assertIsNotNone(first.next)
+        self.assertEqual(first.next.data, 567)
+        self.assertIsNone(first.next.next)
+        second.append_to_tail(LinkedNode(678, None))
+        self.assertIsNotNone(first.next.next)
+        self.assertEqual(first.next.next.data, 678)
+
         # Check ownership
         self.assertEqual(first_meminfo.refcount, 3)
         del second, second_meminfo
@@ -540,7 +557,7 @@ class TestJitClass(TestCase, MemoryLeakMixin):
         def access_dunder(inst):
             return inst.__value
 
-        with self.assertRaises(errors.UntypedAttributeError) as raises:
+        with self.assertRaises(errors.TypingError) as raises:
             access_dunder(inst)
         # It will appear as "_TestJitClass__value" because the `access_dunder`
         # is under the scope of 'TestJitClass'.
@@ -602,6 +619,26 @@ class TestJitClass(TestCase, MemoryLeakMixin):
         self.assertEqual(tc.a, x * y)
         self.assertEqual(tc.b, z)
 
+    def test_generator_method(self):
+        spec = []
+
+        @jitclass(spec)
+        class TestClass(object):
+            def __init__(self):
+                pass
+
+            def gen(self, niter):
+                for i in range(niter):
+                    yield np.arange(i)
+
+        def expected_gen(niter):
+            for i in range(niter):
+                yield np.arange(i)
+
+        for niter in range(10):
+            for expect, got in zip(expected_gen(niter), TestClass().gen(niter)):
+                self.assertPreciseEqual(expect, got)
+
 
 if __name__ == '__main__':
     unittest.main()
