@@ -379,17 +379,17 @@ class TestOperators(TestCase):
     op = LiteralOperatorImpl
 
     _bitwise_opnames = {
-        'bitshift_left_usecase': '<<',
-        'bitshift_ileft_usecase': '<<',
-        'bitshift_right_usecase': '>>',
-        'bitshift_iright_usecase': '>>',
-        'bitwise_and_usecase': '&',
-        'bitwise_iand_usecase': '&',
-        'bitwise_or_usecase': '|',
-        'bitwise_ior_usecase': '|',
-        'bitwise_xor_usecase': '^',
-        'bitwise_ixor_usecase': '^',
-        'bitwise_not_usecase_binary': '~',
+        'bitshift_left_usecase': operator.lshift,
+        'bitshift_ileft_usecase': operator.ilshift,
+        'bitshift_right_usecase': operator.rshift,
+        'bitshift_iright_usecase': operator.irshift,
+        'bitwise_and_usecase': operator.and_,
+        'bitwise_iand_usecase': operator.iand,
+        'bitwise_or_usecase': operator.or_,
+        'bitwise_ior_usecase': operator.ior,
+        'bitwise_xor_usecase': operator.xor,
+        'bitwise_ixor_usecase': operator.ixor,
+        'bitwise_not_usecase_binary': operator.invert,
     }
 
     def run_test_ints(self, pyfunc, x_operands, y_operands, types_list,
@@ -1045,7 +1045,7 @@ class TestOperators(TestCase):
             with self.assertRaises(errors.TypingError, msg=msg) as raises:
                 compile_isolated(pyfunc, argtypes)
             # check error message
-            fmt = 'Invalid usage of {}'
+            fmt = 'Invalid use of {}'
             expecting = fmt.format(opname
                                    if isinstance(opname, str)
                                    else 'Function({})'.format(opname))
@@ -1110,7 +1110,7 @@ class TestOperators(TestCase):
             cfunc = cres.entry_point
             self.assertEqual(pyfunc(val), cfunc(val))
 
-    # XXX test_negate should check for negative and positive zeros and infinites
+    # XXX test_negate should check for negative and positive zeros and infinities
 
     @tag('important')
     def test_negate_npm(self):
@@ -1123,6 +1123,8 @@ class TestOperators(TestCase):
             types.float32,
             types.float64,
             types.complex128,
+            types.boolean,
+            types.boolean,
         ]
         values = [
             1,
@@ -1131,12 +1133,15 @@ class TestOperators(TestCase):
             1.2,
             2.4,
             3.4j,
+            True,
+            False,
         ]
         for ty, val in zip(argtys, values):
             cres = compile_isolated(pyfunc, [ty])
             cfunc = cres.entry_point
             self.assertAlmostEqual(pyfunc(val), cfunc(val))
 
+
     def test_negate(self):
         pyfunc = self.op.negate_usecase
         values = [
@@ -1145,6 +1150,8 @@ class TestOperators(TestCase):
             3,
             1.2,
             3.4j,
+            True,
+            False,
         ]
         cres = compile_isolated(pyfunc, (), flags=force_pyobj_flags)
         cfunc = cres.entry_point
@@ -1161,6 +1168,8 @@ class TestOperators(TestCase):
             types.float32,
             types.float64,
             types.complex128,
+            types.boolean,
+            types.boolean,
         ]
         values = [
             1,
@@ -1169,6 +1178,8 @@ class TestOperators(TestCase):
             1.2,
             2.4,
             3.4j,
+            True,
+            False
         ]
         for ty, val in zip(argtys, values):
             cres = compile_isolated(pyfunc, [ty])
@@ -1183,6 +1194,8 @@ class TestOperators(TestCase):
             3,
             1.2,
             3.4j,
+            True,
+            False,
         ]
         cres = compile_isolated(pyfunc, (), flags=force_pyobj_flags)
         cfunc = cres.entry_point
@@ -1509,6 +1522,37 @@ class TestStaticPower(TestCase):
 
         self._check_pow(exponents, vals)
 
+class TestStringConstComparison(TestCase):
+    """
+    Test comparison of string constants
+    """
+    def test_eq(self):
+        def test_impl1():
+            s = 'test'
+            return s == 'test'
+
+        def test_impl2():
+            s = 'test1'
+            return s == 'test'
+
+        cfunc1 = jit(nopython=True)(test_impl1)
+        cfunc2 = jit(nopython=True)(test_impl2)
+        self.assertEqual(test_impl1(), cfunc1())
+        self.assertEqual(test_impl2(), cfunc2())
+
+    def test_neq(self):
+        def test_impl1():
+            s = 'test'
+            return s != 'test'
+
+        def test_impl2():
+            s = 'test1'
+            return s != 'test'
+
+        cfunc1 = jit(nopython=True)(test_impl1)
+        cfunc2 = jit(nopython=True)(test_impl2)
+        self.assertEqual(test_impl1(), cfunc1())
+        self.assertEqual(test_impl2(), cfunc2())
 
 if __name__ == '__main__':
     unittest.main()
