@@ -14,17 +14,24 @@ Constraints push types forward following the dataflow.
 
 from __future__ import print_function, division, absolute_import
 
+import operator
 import contextlib
 import itertools
 from pprint import pprint
 import traceback
 from collections import OrderedDict
 
-from numba import ir, types, utils, config, six, typing
-from .errors import TypingError, UntypedAttributeError, new_error_context
+from numba import ir, types, utils, config, typing
+from .errors import (TypingError, UntypedAttributeError, new_error_context,
+                     termcolor, UnsupportedError)
 from .funcdesc import qualifying_prefix
 
 
+class NOTSET: pass
+
+# terminal color markup
+_termcolor = termcolor()
+
 class TypeVar(object):
     def __init__(self, context, var):
         self.context = context
@@ -33,6 +40,8 @@ class TypeVar(object):
         self.locked = False
         # Stores source location of first definition
         self.define_loc = None
+        # Qualifiers
+        self.literal_value = NOTSET
 
     def add_type(self, tp, loc):
         assert isinstance(tp, types.Type), type(tp)
@@ -49,10 +58,10 @@ class TypeVar(object):
             if self.type is not None:
                 unified = self.context.unify_pairs(self.type, tp)
                 if unified is None:
-                    msg = "cannot unify %s and %s for '%s', defined at %s"
+                    msg = "Cannot unify %s and %s for '%s', defined at %s"
                     raise TypingError(msg % (self.type, tp, self.var,
                                              self.define_loc),
-                                      loc=loc)
+                                      loc=self.define_loc)
             else:
                 # First time definition
                 unified = tp
@@ -62,21 +71,22 @@ class TypeVar(object):
 
         return self.type
 
-    def lock(self, tp, loc):
+    def lock(self, tp, loc, literal_value=NOTSET):
         assert isinstance(tp, types.Type), type(tp)
         assert not self.locked
 
         # If there is already a type, ensure we can convert it to the
         # locked type.
         if (self.type is not None and
-            self.context.can_convert(self.type, tp) is None):
+                self.context.can_convert(self.type, tp) is None):
             raise TypingError("No conversion from %s to %s for "
-                              "'%s'" % (tp, self.type, self.var))
+                              "'%s'" % (tp, self.type, self.var), loc=loc)
 
         self.type = tp
         self.locked = True
         if self.define_loc is None:
             self.define_loc = loc
+        self.literal_value = literal_value
 
     def union(self, other, loc):
         if other.type is not None:
@@ -85,7 +95,7 @@ class TypeVar(object):
         return self.type
 
     def __repr__(self):
-        return '%s := %s' % (self.var, self.type)
+        return '%s := %s' % (self.var, self.type or "<undecided>")
 
     @property
     def defined(self):
@@ -95,7 +105,8 @@ class TypeVar(object):
         return (self.type,) if self.type is not None else ()
 
     def getone(self):
-        assert self.type is not None
+        if self.type is None:
+            raise TypingError("Undecided type {}".format(self))
         return self.type
 
     def __len__(self):
@@ -118,7 +129,8 @@ class ConstraintNetwork(object):
         """
         Execute all constraints.  Errors are caught and returned as a list.
         This allows progressing even though some constraints may fail
-        due to lack of information (e.g. imprecise types such as List(undefined)).
+        due to lack of information
+        (e.g. imprecise types such as List(undefined)).
         """
         errors = []
         for constraint in self.constraints:
@@ -128,13 +140,17 @@ class ConstraintNetwork(object):
                 try:
                     constraint(typeinfer)
                 except TypingError as e:
+                    e = TypingError(str(e),
+                                    loc=constraint.loc,
+                                    highlighting=False)
                     errors.append(e)
                 except Exception:
                     msg = "Internal error at {con}:\n{sep}\n{err}{sep}\n"
                     e = TypingError(msg.format(con=constraint,
                                                err=traceback.format_exc(),
-                                               sep='--%<' +'-' * 65),
-                                    loc=constraint.loc)
+                                               sep='--%<' + '-' * 76),
+                                    loc=constraint.loc,
+                                    highlighting=False)
                     errors.append(e)
         return errors
 
@@ -150,13 +166,15 @@ class Propagate(object):
         self.loc = loc
 
     def __call__(self, typeinfer):
-        with new_error_context("typing of assignment at {0}", self.loc):
+        with new_error_context("typing of assignment at {0}", self.loc,
+                               loc=self.loc):
             typeinfer.copy_type(self.src, self.dst, loc=self.loc)
             # If `dst` is refined, notify us
             typeinfer.refine_map[self.dst] = self
 
     def refine(self, typeinfer, target_type):
         # Do not back-propagate to locked variables (e.g. constants)
+        assert target_type.is_precise()
         typeinfer.add_type(self.src, target_type, unless_locked=True,
                            loc=self.loc)
 
@@ -177,6 +195,7 @@ class ArgConstraint(object):
             ty = src.getone()
             if isinstance(ty, types.Omitted):
                 ty = typeinfer.context.resolve_value_type(ty.value)
+            assert ty.is_precise()
             typeinfer.add_type(self.dst, ty, loc=self.loc)
 
 
@@ -190,13 +209,13 @@ class BuildTupleConstraint(object):
         with new_error_context("typing of tuple at {0}", self.loc):
             typevars = typeinfer.typevars
             tsets = [typevars[i.name].get() for i in self.items]
-            oset = typevars[self.target]
             for vals in itertools.product(*tsets):
                 if vals and all(vals[0] == v for v in vals):
                     tup = types.UniTuple(dtype=vals[0], count=len(vals))
                 else:
                     # empty tuples fall here as well
                     tup = types.Tuple(vals)
+                assert tup.is_precise()
                 typeinfer.add_type(self.target, tup, loc=self.loc)
 
 
@@ -210,7 +229,6 @@ class _BuildContainerConstraint(object):
     def __call__(self, typeinfer):
         with new_error_context("typing of list at {0}", self.loc):
             typevars = typeinfer.typevars
-            oset = typevars[self.target]
             tsets = [typevars[i.name].get() for i in self.items]
             if not tsets:
                 typeinfer.add_type(self.target,
@@ -233,7 +251,6 @@ class BuildSetConstraint(_BuildContainerConstraint):
     container_type = types.Set
 
 
-
 class ExhaustIterConstraint(object):
     def __init__(self, target, count, iterator, loc):
         self.target = target
@@ -244,25 +261,28 @@ class ExhaustIterConstraint(object):
     def __call__(self, typeinfer):
         with new_error_context("typing of exhaust iter at {0}", self.loc):
             typevars = typeinfer.typevars
-            oset = typevars[self.target]
             for tp in typevars[self.iterator.name].get():
                 # unpack optional
                 tp = tp.type if isinstance(tp, types.Optional) else tp
                 if isinstance(tp, types.BaseTuple):
                     if len(tp) == self.count:
+                        assert tp.is_precise()
                         typeinfer.add_type(self.target, tp, loc=self.loc)
                         break
                     else:
                         raise ValueError("wrong tuple length for %r: "
                                          "expected %d, got %d"
-                                         % (self.iterator.name, self.count, len(tp)))
+                                         % (self.iterator.name, self.count,
+                                            len(tp)))
                 elif isinstance(tp, types.IterableType):
                     tup = types.UniTuple(dtype=tp.iterator_type.yield_type,
                                          count=self.count)
+                    assert tup.is_precise()
                     typeinfer.add_type(self.target, tup, loc=self.loc)
                     break
             else:
-                raise TypingError("failed to unpack {}".format(tp), loc=self.loc)
+                raise TypingError("failed to unpack {}".format(tp),
+                                  loc=self.loc)
 
 
 class PairFirstConstraint(object):
@@ -274,11 +294,11 @@ class PairFirstConstraint(object):
     def __call__(self, typeinfer):
         with new_error_context("typing of pair-first at {0}", self.loc):
             typevars = typeinfer.typevars
-            oset = typevars[self.target]
             for tp in typevars[self.pair.name].get():
                 if not isinstance(tp, types.Pair):
                     # XXX is this an error?
                     continue
+                assert tp.first_type.is_precise()
                 typeinfer.add_type(self.target, tp.first_type, loc=self.loc)
 
 
@@ -291,11 +311,11 @@ class PairSecondConstraint(object):
     def __call__(self, typeinfer):
         with new_error_context("typing of pair-second at {0}", self.loc):
             typevars = typeinfer.typevars
-            oset = typevars[self.target]
             for tp in typevars[self.pair.name].get():
                 if not isinstance(tp, types.Pair):
                     # XXX is this an error?
                     continue
+                assert tp.second_type.is_precise()
                 typeinfer.add_type(self.target, tp.second_type, loc=self.loc)
 
 
@@ -305,7 +325,7 @@ class StaticGetItemConstraint(object):
         self.value = value
         self.index = index
         if index_var is not None:
-            self.fallback = IntrinsicCallConstraint(target, 'getitem',
+            self.fallback = IntrinsicCallConstraint(target, operator.getitem,
                                                     (value, index_var), {},
                                                     None, loc)
         else:
@@ -315,11 +335,11 @@ class StaticGetItemConstraint(object):
     def __call__(self, typeinfer):
         with new_error_context("typing of static-get-item at {0}", self.loc):
             typevars = typeinfer.typevars
-            oset = typevars[self.target]
             for ty in typevars[self.value.name].get():
-                itemty = typeinfer.context.resolve_static_getitem(value=ty,
-                                                                  index=self.index)
+                itemty = typeinfer.context.resolve_static_getitem(
+                            value=ty, index=self.index)
                 if itemty is not None:
+                    assert itemty.is_precise()
                     typeinfer.add_type(self.target, itemty, loc=self.loc)
                 elif self.fallback is not None:
                     self.fallback(typeinfer)
@@ -345,19 +365,38 @@ def fold_arg_vars(typevars, args, vararg, kws):
         return
 
     args = tuple(a.getone() for a in argtypes)
+
     pos_args = args[:n_pos_args]
     if vararg is not None:
-        if not isinstance(args[-1], types.BaseTuple):
+        errmsg = "*args in function call should be a tuple, got %s"
+        # Handle constant literal used for `*args`
+        if isinstance(args[-1], types.Literal):
+            const_val = args[-1].literal_value
+            # Is the constant value a tuple?
+            if not isinstance(const_val, tuple):
+                raise TypeError(errmsg % (args[-1],))
+            # Append the elements in the const tuple to the positional args
+            pos_args += const_val
+        # Handle non-constant
+        elif not isinstance(args[-1], types.BaseTuple):
             # Unsuitable for *args
             # (Python is more lenient and accepts all iterables)
-            raise TypeError("*args in function call should be a tuple, got %s"
-                            % (args[-1],))
-        pos_args += args[-1].types
+            raise TypeError(errmsg % (args[-1],))
+        else:
+            # Append the elements in the tuple to the positional args
+            pos_args += args[-1].types
+        # Drop the last arg
         args = args[:-1]
     kw_args = dict(zip(kwds, args[n_pos_args:]))
     return pos_args, kw_args
 
 
+def _is_array_not_precise(arrty):
+    """Check type is array and it is not precise
+    """
+    return isinstance(arrty, types.Array) and not arrty.is_precise()
+
+
 class CallConstraint(object):
     """Constraint for calling functions.
     Perform case analysis foreach combinations of argument types.
@@ -373,9 +412,12 @@ class CallConstraint(object):
         self.loc = loc
 
     def __call__(self, typeinfer):
-        with new_error_context("typing of call at {0}", self.loc):
+        msg = "typing of call at {0}\n".format(self.loc)
+        with new_error_context(msg):
             typevars = typeinfer.typevars
-            fnty = typevars[self.func].getone()
+            with new_error_context(
+                    "resolving caller type: {}".format(self.func)):
+                fnty = typevars[self.func].getone()
             with new_error_context("resolving callee type: {0}", fnty):
                 self.resolve(typeinfer, typevars, fnty)
 
@@ -389,25 +431,39 @@ class CallConstraint(object):
             return
         pos_args, kw_args = r
 
+        # Check argument to be precise
+        for a in itertools.chain(pos_args, kw_args.values()):
+            if not a.is_precise():
+                # Getitem on non-precise array is allowed to
+                # support array-comprehension
+                if fnty == operator.getitem and isinstance(pos_args[0], types.Array):
+                    pass
+                # Otherwise, don't compute type yet
+                else:
+                    return
+
         # Resolve call type
         sig = typeinfer.resolve_call(fnty, pos_args, kw_args)
+
         if sig is None:
+            # Note: duplicated error checking.
+            #       See types.BaseFunction.get_call_type
             # Arguments are invalid => explain why
-            headtemp = "Invalid usage of {0} with parameters ({1})"
+            headtemp = "Invalid use of {0} with parameters ({1})"
             args = [str(a) for a in pos_args]
             args += ["%s=%s" % (k, v) for k, v in sorted(kw_args.items())]
             head = headtemp.format(fnty, ', '.join(map(str, args)))
             desc = context.explain_function_type(fnty)
             msg = '\n'.join([head, desc])
-            raise TypingError(msg, loc=self.loc)
+            raise TypingError(msg)
 
         typeinfer.add_type(self.target, sig.return_type, loc=self.loc)
 
         # If the function is a bound function and its receiver type
         # was refined, propagate it.
         if (isinstance(fnty, types.BoundFunction)
-            and sig.recvr is not None
-            and sig.recvr != fnty.this):
+                and sig.recvr is not None
+                and sig.recvr != fnty.this):
             refined_this = context.unify_pairs(sig.recvr, fnty.this)
             if refined_this is not None and refined_this.is_precise():
                 refined_fnty = fnty.copy(this=refined_this)
@@ -424,10 +480,26 @@ class CallConstraint(object):
             if target.defined:
                 targetty = target.getone()
                 if context.unify_pairs(targetty, sig.return_type) == targetty:
-                    sig.return_type = targetty
+                    sig = sig.replace(return_type=targetty)
 
         self.signature = sig
 
+        target_type = typevars[self.target].getone()
+        if (isinstance(target_type, types.Array)
+                and isinstance(sig.return_type.dtype, types.Undefined)):
+            typeinfer.refine_map[self.target] = self
+
+    def refine(self, typeinfer, updated_type):
+        # Is getitem?
+        if self.func == operator.getitem:
+            aryty = typeinfer.typevars[self.args[0].name].getone()
+            # is array not precise?
+            if _is_array_not_precise(aryty):
+                # allow refinement of dtype
+                assert updated_type.is_precise()
+                newtype = aryty.copy(dtype=updated_type.dtype)
+                typeinfer.add_type(self.args[0].name, newtype, loc=self.loc)
+
     def get_call_signature(self):
         return self.signature
 
@@ -435,7 +507,10 @@ class CallConstraint(object):
 class IntrinsicCallConstraint(CallConstraint):
     def __call__(self, typeinfer):
         with new_error_context("typing of intrinsic-call at {0}", self.loc):
-            self.resolve(typeinfer, typeinfer.typevars, fnty=self.func)
+            fnty = self.func
+            if fnty in utils.OPERATORS_TO_BUILTINS:
+                fnty = typeinfer.resolve_value_type(None, fnty)
+            self.resolve(typeinfer, typeinfer.typevars, fnty=fnty)
 
 
 class GetAttrConstraint(object):
@@ -453,14 +528,17 @@ class GetAttrConstraint(object):
             for ty in valtys:
                 attrty = typeinfer.context.resolve_getattr(ty, self.attr)
                 if attrty is None:
-                    raise UntypedAttributeError(ty, self.attr, loc=self.inst.loc)
+                    raise UntypedAttributeError(ty, self.attr,
+                                                loc=self.inst.loc)
                 else:
+                    assert attrty.is_precise()
                     typeinfer.add_type(self.target, attrty, loc=self.loc)
             typeinfer.refine_map[self.target] = self
 
     def refine(self, typeinfer, target_type):
         if isinstance(target_type, types.BoundFunction):
             recvr = target_type.this
+            assert recvr.is_precise()
             typeinfer.add_type(self.value.name, recvr, loc=self.loc)
             source_constraint = typeinfer.refine_map.get(self.value.name)
             if source_constraint is not None:
@@ -492,6 +570,12 @@ class SetItemConstraint(object):
             if sig is None:
                 raise TypingError("Cannot resolve setitem: %s[%s] = %s" %
                                   (targetty, idxty, valty), loc=self.loc)
+
+            # For array setitem, refine imprecise array dtype
+            if _is_array_not_precise(targetty):
+                assert sig.args[0].is_precise()
+                typeinfer.add_type(self.target.name, sig.args[0], loc=self.loc)
+
             self.signature = sig
 
     def get_call_signature(self):
@@ -516,7 +600,8 @@ class StaticSetItemConstraint(object):
             idxty = typevars[self.index_var.name].getone()
             valty = typevars[self.value.name].getone()
 
-            sig = typeinfer.context.resolve_static_setitem(targetty, self.index, valty)
+            sig = typeinfer.context.resolve_static_setitem(targetty,
+                                                           self.index, valty)
             if sig is None:
                 sig = typeinfer.context.resolve_setitem(targetty, idxty, valty)
             if sig is None:
@@ -625,6 +710,7 @@ class TypeVarMap(dict):
 # A temporary mapping of {function name: dispatcher object}
 _temporary_dispatcher_map = {}
 
+
 @contextlib.contextmanager
 def register_dispatcher(disp):
     """
@@ -645,6 +731,7 @@ def register_dispatcher(disp):
 
 typeinfer_extensions = {}
 
+
 class TypeInferer(object):
     """
     Operates on block that shares the same ir.Scope.
@@ -667,7 +754,7 @@ class TypeInferer(object):
 
         # { index: mangled name }
         self.arg_names = {}
-        #self.return_type = None
+        # self.return_type = None
         # Set of assumed immutable globals
         self.assumed_immutables = set()
         # Track all calls and associated constraints
@@ -748,7 +835,7 @@ class TypeInferer(object):
             if retvar.name in cloned.typevars:
                 typevar = cloned.typevars[retvar.name]
                 if typevar and typevar.defined:
-                    rettypes.add(typevar.getone())
+                    rettypes.add(types.unliteral(typevar.getone()))
         if not rettypes:
             return
         # unify return types
@@ -785,14 +872,15 @@ class TypeInferer(object):
             self.propagate_refined_type(var, unified)
 
     def add_calltype(self, inst, signature):
+        assert signature is not None
         self.calltypes[inst] = signature
 
     def copy_type(self, src_var, dest_var, loc):
-        unified = self.typevars[dest_var].union(self.typevars[src_var], loc=loc)
+        self.typevars[dest_var].union(self.typevars[src_var], loc=loc)
 
-    def lock_type(self, var, tp, loc):
+    def lock_type(self, var, tp, loc, literal_value=NOTSET):
         tv = self.typevars[var]
-        tv.lock(tp, loc=loc)
+        tv.lock(tp, loc=loc, literal_value=literal_value)
 
     def propagate_refined_type(self, updated_var, updated_type):
         source_constraint = self.refine_map.get(updated_var)
@@ -806,14 +894,82 @@ class TypeInferer(object):
         """
         typdict = utils.UniqueDict()
 
+        def find_offender(name, exhaustive=False):
+            # finds the offending variable definition by name
+            # if exhaustive is set it will try and trace through temporary
+            # variables to find a concrete offending definition.
+            offender = None
+            for block in self.func_ir.blocks.values():
+                offender = block.find_variable_assignment(name)
+                if offender is not None:
+                    if not exhaustive:
+                        break
+                    try: # simple assignment
+                        hasattr(offender.value, 'name')
+                        offender_value = offender.value.name
+                    except (AttributeError, KeyError):
+                        break
+                    orig_offender = offender
+                    if offender_value.startswith('$'):
+                        offender = find_offender(offender_value,
+                                                 exhaustive=exhaustive)
+                        if offender is None:
+                            offender = orig_offender
+                    break
+            return offender
+
+        def diagnose_imprecision(offender):
+            # helper for diagnosing imprecise types
+
+            list_msg = """\n
+For Numba to be able to compile a list, the list must have a known and
+precise type that can be inferred from the other variables. Whilst sometimes
+the type of empty lists can be inferred, this is not always the case, see this
+documentation for help:
+
+http://numba.pydata.org/numba-doc/latest/user/troubleshoot.html#my-code-has-an-untyped-list-problem
+"""
+            if offender is not None:
+                # This block deals with imprecise lists
+                if hasattr(offender, 'value'):
+                    if hasattr(offender.value, 'op'):
+                        # might be `foo = []`
+                        if offender.value.op == 'build_list':
+                            return list_msg
+                        # or might be `foo = list()`
+                        elif offender.value.op == 'call':
+                            try: # assignment involving a call
+                                call_name = offender.value.func.name
+                                # find the offender based on the call name
+                                offender = find_offender(call_name)
+                                if isinstance(offender.value, ir.Global):
+                                    if offender.value.name == 'list':
+                                        return list_msg
+                            except (AttributeError, KeyError):
+                                pass
+            return "" # no help possible
+
         def check_var(name):
             tv = self.typevars[name]
             if not tv.defined:
-                raise TypingError("Undefined variable '%s'" % (var,))
+                offender = find_offender(name)
+                val = getattr(offender, 'value', 'unknown operation')
+                loc = getattr(offender, 'loc', ir.unknown_loc)
+                msg = ("Type of variable '%s' cannot be determined, operation:"
+                      " %s, location: %s")
+                raise TypingError(msg % (var, val, loc), loc)
             tp = tv.getone()
             if not tp.is_precise():
-                raise TypingError("Can't infer type of variable '%s': %s" % (var, tp))
-            typdict[var] = tp
+                offender = find_offender(name, exhaustive=True)
+                msg = ("Cannot infer the type of variable '%s'%s, "
+                      "have imprecise type: %s. %s")
+                istmp = " (temporary variable)" if var.startswith('$') else ""
+                loc = getattr(offender, 'loc', ir.unknown_loc)
+                # is this an untyped list? try and provide help
+                extra_msg = diagnose_imprecision(offender)
+                raise TypingError(msg % (var, istmp, tp, extra_msg), loc)
+            else: # type is precise, hold it
+                typdict[var] = tp
 
         # For better error display, check first user-visible vars, then
         # temporaries
@@ -839,13 +995,15 @@ class TypeInferer(object):
         for index, name in self.arg_names.items():
             arg_types[index] = typdict[name]
         state_types = [typdict[var_name] for var_name in gi.state_vars]
-        yield_types = [typdict[y.inst.value.name] for y in gi.get_yield_points()]
+        yield_types = [typdict[y.inst.value.name]
+                       for y in gi.get_yield_points()]
         if not yield_types:
-            raise TypingError("Cannot type generator: it does not yield any value")
+            msg = "Cannot type generator: it does not yield any value"
+            raise TypingError(msg)
         yield_type = self.context.unify_types(*yield_types)
         if yield_type is None:
-            raise TypingError("Cannot type generator: cannot unify yielded types "
-                              "%s" % (yield_types,))
+            msg = "Cannot type generator: cannot unify yielded types %s"
+            raise TypingError(msg % (yield_types,))
         return types.Generator(self.func_id.func, yield_type, arg_types,
                                state_types, has_finalizer=True)
 
@@ -863,9 +1021,39 @@ class TypeInferer(object):
         if rettypes:
             unified = self.context.unify_types(*rettypes)
             if unified is None or not unified.is_precise():
+                def check_type(atype):
+                    lst = []
+                    for k, v in self.typevars.items():
+                        if atype == v.type:
+                            lst.append(k)
+                    returns = {}
+                    for x in reversed(lst):
+                        for block in self.func_ir.blocks.values():
+                            for instr in block.find_insts(ir.Return):
+                                value = instr.value
+                                if isinstance(value, ir.Var):
+                                    name = value.name
+                                else:
+                                    pass
+                                if x == name:
+                                    returns[x] = instr
+                                    break
+
+                    for name, offender in returns.items():
+                        loc = getattr(offender, 'loc', ir.unknown_loc)
+                        msg = ("Return of: IR name '%s', type '%s', "
+                               "location: %s")
+                        interped = msg % (name, atype, loc.strformat())
+                    return interped
+
+                problem_str = []
+                for xtype in rettypes:
+                    problem_str.append(_termcolor.errmsg(check_type(xtype)))
+
                 raise TypingError("Can't unify return type from the "
                                   "following types: %s"
-                                  % ", ".join(sorted(map(str, rettypes))))
+                                  % ", ".join(sorted(map(str, rettypes))) +
+                                  "\n" + "\n".join(problem_str))
             return unified
         else:
             # Function without a successful return path
@@ -905,7 +1093,8 @@ class TypeInferer(object):
             f = typeinfer_extensions[type(inst)]
             f(inst, self)
         else:
-            raise NotImplementedError(inst)
+            msg = "Unsupported constraint encountered: %s" % inst
+            raise UnsupportedError(msg, loc=inst.loc)
 
     def typeof_setitem(self, inst):
         constraint = SetItemConstraint(target=inst.target, index=inst.index,
@@ -955,7 +1144,9 @@ class TypeInferer(object):
         elif isinstance(value, ir.Yield):
             self.typeof_yield(inst, inst.target, value)
         else:
-            raise NotImplementedError(type(value), str(value))
+            msg = ("Unsupported assignment encountered: %s %s" %
+                   (type(value), str(value)))
+            raise UnsupportedError(msg, loc=inst.loc)
 
     def resolve_value_type(self, inst, val):
         """
@@ -976,10 +1167,11 @@ class TypeInferer(object):
 
     def typeof_const(self, inst, target, const):
         ty = self.resolve_value_type(inst, const)
-        # Special case string constant as Const type
-        if ty == types.string:
-            ty = types.Const(value=const)
-        self.lock_type(target.name, ty, loc=inst.loc)
+        if inst.value.use_literal_type:
+            lit = types.maybe_literal(value=const)
+        else:
+            lit = None
+        self.add_type(target.name, lit or ty, loc=inst.loc)
 
     def typeof_yield(self, inst, target, yield_):
         # Sending values into generators isn't supported.
@@ -990,7 +1182,7 @@ class TypeInferer(object):
         Ensure that builtins are not modified.
         """
         if (gvar.name in ('range', 'xrange') and
-            gvar.value not in utils.RANGE_ITER_OBJECTS):
+                gvar.value not in utils.RANGE_ITER_OBJECTS):
             bad = True
         elif gvar.name == 'slice' and gvar.value is not slice:
             bad = True
@@ -1044,18 +1236,19 @@ class TypeInferer(object):
             typ = self.resolve_value_type(inst, gvar.value)
         except TypingError as e:
             if (gvar.name == self.func_id.func_name
-                and gvar.name in _temporary_dispatcher_map):
+                    and gvar.name in _temporary_dispatcher_map):
                 # Self-recursion case where the dispatcher is not (yet?) known
                 # as a global variable
                 typ = types.Dispatcher(_temporary_dispatcher_map[gvar.name])
             else:
-                e.patch_message("Untyped global name '%s': %s"
-                                % (gvar.name, e))
+                msg = _termcolor.errmsg("Untyped global name '%s':") + " %s"
+                e.patch_message(msg % (gvar.name, e))
                 raise
 
         if isinstance(typ, types.Dispatcher) and typ.dispatcher.is_compiling:
             # Recursive call
-            callframe = self.context.callstack.findfirst(typ.dispatcher.py_func)
+            callstack = self.context.callstack
+            callframe = callstack.findfirst(typ.dispatcher.py_func)
             if callframe is not None:
                 typ = types.RecursiveCall(typ)
             else:
@@ -1068,7 +1261,10 @@ class TypeInferer(object):
             typ = typ.copy(readonly=True)
 
         self.sentry_modified_builtin(inst, gvar)
-        self.lock_type(target.name, typ, loc=inst.loc)
+        # Setting literal_value for globals because they are handled
+        # like const value in numba
+        lit = types.maybe_literal(gvar.value)
+        self.lock_type(target.name, lit or typ, loc=inst.loc)
         self.assumed_immutables.add(inst)
 
     def typeof_expr(self, inst, target, expr):
@@ -1095,7 +1291,8 @@ class TypeInferer(object):
                                               loc=expr.loc)
             self.constraints.append(constraint)
         elif expr.op == 'binop':
-            self.typeof_intrinsic_call(inst, target, expr.fn, expr.lhs, expr.rhs)
+            self.typeof_intrinsic_call(inst, target, expr.fn, expr.lhs,
+                                       expr.rhs)
         elif expr.op == 'inplace_binop':
             self.typeof_intrinsic_call(inst, target, expr.fn,
                                        expr.lhs, expr.rhs)
@@ -1109,7 +1306,9 @@ class TypeInferer(object):
             self.constraints.append(constraint)
             self.calls.append((inst.value, constraint))
         elif expr.op == 'getitem':
-            self.typeof_intrinsic_call(inst, target, 'getitem', expr.value, expr.index)
+            self.typeof_intrinsic_call(
+                inst, target, operator.getitem, expr.value, expr.index,
+                )
         elif expr.op == 'getattr':
             constraint = GetAttrConstraint(target.name, attr=expr.attr,
                                            value=expr.value, loc=inst.loc,
@@ -1131,8 +1330,12 @@ class TypeInferer(object):
             self.constraints.append(Propagate(dst=target.name,
                                               src=expr.value.name,
                                               loc=inst.loc))
+        elif expr.op == 'make_function':
+            self.lock_type(target.name, types.MakeFunctionLiteral(expr),
+                           loc=inst.loc, literal_value=expr)
         else:
-            raise NotImplementedError(type(expr), expr)
+            msg = "Unsupported op-code encountered: %s" % expr
+            raise UnsupportedError(msg, loc=inst.loc)
 
     def typeof_call(self, inst, target, call):
         constraint = CallConstraint(target.name, call.func.name, call.args,
