@@ -4,7 +4,6 @@ Define typing templates
 from __future__ import print_function, division, absolute_import
 
 import functools
-from functools import reduce
 import operator
 import sys
 from types import MethodType
@@ -31,6 +30,17 @@ class Signature(object):
         self.recvr = recvr
         self.pysig = pysig
 
+    def replace(self, **kwargs):
+        """Copy and replace the given attributes provided as keyword arguments.
+        Returns an updated copy.
+        """
+        curstate = dict(return_type=self.return_type,
+                        args=self.args,
+                        recvr=self.recvr,
+                        pysig=self.pysig)
+        curstate.update(kwargs)
+        return Signature(**curstate)
+
     def __getstate__(self):
         """
         Needed because of __slots__.
@@ -156,6 +166,9 @@ class FunctionTemplate(object):
     # subclass overide-able
     unsafe_casting = True
 
+    # Whether the typing support literals
+    support_literals = False
+
     def __init__(self, context):
         self.context = context
 
@@ -287,7 +300,7 @@ class _OverloadFunctionTemplate(AbstractTemplate):
         Type the overloaded function by compiling the appropriate
         implementation for the given args.
         """
-        cache_key = self.context, args, tuple(kws.items())
+        cache_key = self.context, tuple(args), tuple(kws.items())
         try:
             disp = self._impl_cache[cache_key]
         except KeyError:
@@ -527,7 +540,7 @@ class _OverloadMethodTemplate(_OverloadAttributeTemplate):
         class MethodTemplate(AbstractTemplate):
             key = (self.key, attr)
             def generic(_, args, kws):
-                args = (typ,) + args
+                args = (typ,) + tuple(args)
                 sig = self._resolve_impl_sig(typ, attr, args, kws)
                 if sig is not None:
                     return sig.as_method()
@@ -559,7 +572,7 @@ def make_overload_method_template(typ, attr, overload_func):
                                             base=_OverloadMethodTemplate)
 
 
-def bound_function(template_key):
+def bound_function(template_key, support_literals=False):
     """
     Wrap an AttributeTemplate resolve_* method to allow it to
     resolve an instance method's signature rather than a instance attribute.
@@ -587,6 +600,7 @@ def bound_function(template_key):
                         sig.recvr = ty
                     return sig
 
+            MethodTemplate.support_literals = support_literals
             return types.BoundFunction(MethodTemplate, ty)
         return attribute_resolver
     return wrapper
